#!/usr/bin/python3

def isInnerEdge(G, e):
    """
    Devolve se a semi-aresta e é interna
    """
    return e.label is None

def isOuterEdge(G, e):
    """
    Devolve se a semi-aresta e é externa (faz parte de C_o)
    """
    return e.label is not None

def BadCycle(G, corners, first, kind):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, uma semi-aresta first e um número kind indicando a relação
    entre a posição de first e do ciclo ruim C encontrado na borda ou no canto de H
    
    kind = 0: a próxima semi-aresta interna a partir de first é a primeira
    semi-aresta de Q_{AH}(C)
    kind = 1: a primeira semi-aresta interna anterior a first é a última
    semi-aresta de Q_{AH}(C)
    kind = 2: a face incidente a first contém o Q_{AH}(C) de um ciclo ruim com
    n_{AH}(C) = 0
    kind = 3: first é a semi-aresta que forma a perna anti-horária de um ciclo
    de borda ruim e first.origin está no contorno de C
    
    Devolve uma lista de semi-arestas que formam um ciclo ruim
    """
    e = first
    cycle = []
    if kind == 0: # o próximo ciclo é um ciclo com n_{AH}(C) = 0
        while isOuterEdge(G, e): e = e.next_edge
        v0 = e.origin
        cycle += [e]
        e = e.next_edge
        while isInnerEdge(G, e):
            cycle += [e]
            e = e.next_edge
        e = e.outer_next
        while e.destination != v0:
            cycle += [e.twin]
            e = e.outer_next
    elif kind == 1: # o ciclo anterior é um ciclo com n_{AH}(C) = 0
        while isOuterEdge(G, e): e = e.prev_edge
        while isInnerEdge(G, e): e = e.prev_edge
        return BadCycle(G, corners, e, 0)
    elif kind == 2: # a face incidente a first contém o Q_{AH}(C) de um ciclo ruim
        return BadCycle(G, corners, findBadCycle(G, corners, first), 0)
    elif kind == 3: # first é a perna anti-horária de um ciclo de borda ruim
        e = first.prev_edge
        while isInnerEdge(G, e):
            cycle += [e]
            e = e.prev_edge
        v0 = e.destination
        cycle = list(reversed(cycle))
        e = first.twin.next_edge
        while isInnerEdge(G, e):
            cycle += [e]
            e = e.next_edge
        e = e.outer_next
        while e.destination != v0:
            cycle += [e.twin]
            e = e.outer_next
    return cycle

def findBadCycle(G, corners, first):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma semi-aresta first tal que a face incidente a first
    contém todo o Q_{AH}(C) e que contém menos que 2 cantos de H
    
    Percorre a face f incidente a first verificando o número de cantos que cada
    ciclo formado entre f e C_o(H) contém
    Se o número de cantos for menor que 2, o ciclo é necessariamente um canto ou
    borda ruim
    
    Devolve uma semi-aresta e tal que a próxima semi-aresta interna a partir de e
    é a primeira semi-aresta de Q_{AH} do ciclo ruim
    """
    f = first.incident_face
    e = first.prev_edge
    while e != first:
        corners_count = 0
        while isOuterEdge(G, e): e = e.prev_edge
        e = e.next_edge.outer_next
        while e.incident_face != f:
            if e.destination in corners: corners_count += 1
            e = e.outer_next
        if corners_count < 2: return e
    raise Exception("Bad Cycle not found.")

def BadCycleBetween(G, e, f, g):
    """
    Recebe um PlaneGraph G e uma semi-aresta e no contorno de um ciclo interno
    ruim entre as faces separadoras f e g
    
    Devolve uma lista de semi-arestas formando o contorno do ciclo ruim
    """
    cycle = []
    if e.twin.incident_face == g:
        e = e.twin.prev_edge
        while e.twin.incident_face != f:
            cycle += [e.twin]
            e = e.prev_edge
        e = e.twin.prev_edge
        while e.twin.incident_face != g:
            cycle += [e.twin]
            e = e.prev_edge
    else:
        e = e.next_edge.twin.next_edge
        while e.twin.incident_face != f:
            cycle += [e]
            e = e.next_edge
        e = e.twin.next_edge
        while e.twin.incident_face != g:
            cycle += [e.twin]
            e = e.next_edge
    return cycle

def BadBigCycle(G, leg, first):
    """
    Recebe um PlaneGraph G, a perna horária leg e uma semi-aresta first incidente
    ao vértice anterior de C, onde C é um ciclo crítico maximal anexado a um NS-
    caminho mais à esquerda P de um subgrafo de G
    
    Devolve uma lista de semi-arestas formando o contorno do ciclo interno ruim C
    """
    cycle = []
    e = first.next_edge
    while e != leg:
        cycle += [e]
        e = e.next_edge
    e = e.twin.next_edge
    while not e.is_on_p: 
        cycle += [e]
        e = e.next_edge
    e = e.twin.next_edge
    while e.twin != first:
        cycle += [e]
        e = e.next_edge
    return cycle

def BadCycleInside(G, inner_leg, outer_leg, first):
    """
    Recebe um PlaneGraph G, as pernas inner_leg e outer_leg de um ciclo crítico
    C' anexado a Q_{AH}(C), onde C é um ciclo crítico anexado a um NS-caminho
    mais à esquerda P de um subgrafo de G, e uma semi-aresta first
    
    - inner_leg é a semi-aresta de C' contida em C se C' tiver 3 pernas, senão
    é None
    - outer_leg é a perna horária de C
    - first é a perna incidente ao vértice anterior de C'
    
    Devolve uma lista de semi-arestas formando o contorno do ciclo interno ruim C'
    """
    cycle = []
    e = first.next_edge
    if inner_leg:
        while e.twin != inner_leg:
            cycle += [e]
            e = e.next_edge
        e = e.twin.next_edge
    while e.order != e.twin.order:
        cycle += [e]
        e = e.next_edge
    e = e.twin.next_edge
    while e != outer_leg and e.twin != first:
        cycle += [e]
        e = e.next_edge
    if e == outer_leg: e = e.twin.next_edge
    while e.twin != first:
        cycle += [e]
        e = e.next_edge
    return cycle
