#!/usr/bin/python3

import sys
from Utilities import intoargs

class PlaneGraph(object):
    """
    Implementação de grafo plano como DCEL (doubly connected edge list), a ser
    utilizada pelos algoritmos RectangularDrawing e RectangularGridDrawing

    Para o uso nos algoritmos, o contorno das faces deve ser dado em sentido anti-horário
    
    Na construção de um grafo plano, são recebidos os parâmetros:
    - vertices: uma lista de tuplas (nome, coords) em que nome é uma string ou None,
        servindo de nome a cada vértice, e coords é uma tupla (x, y) de floats ou None
    - edges: uma lista de tuplas (nome, (i, j)) em que nome é uma string ou None,
        servindo de nome a cada aresta, e i, j são os índices dos vértices da aresta;
        cada aresta gera duas semi-arestas, de i a j e de j a i, sendo a
        semi-aresta de i a j considerada a semi-aresta de sentido padrão
    - faces*: uma lista de tuplas (nome, contorno) em que nome é uma string ou None,
        servindo de nome a cada face, e contorno é uma tupla (orientação, arestas);
        arestas é uma lista de índices das arestas formando o ciclo de contorno da
        face e a orientação assume o valor '+' ou '-' para indicar se quando o
        contorno da face é percorrido no sentido anti-horário, arestas[0] é
        percorrida no sentido padrão ('+') ou não ('-')
        
    *a face exterior do grafo plano deve ser sempre a primeira face na lista
    
    Os elementos (vértices, arestas e faces) do PlaneGraph G ficam armazenados nas
    listas G.vertices, G.edges e G.faces, respectivamente, na ordem que foram
    recebidos pelo construtor.
    
    Para acessar o vértice de índice k, pode ser utilizado G.vertex(k)
    Para acessar a face de índice k, pode ser utilizado G.face(k)
    
    Cada elemento de G.edges é uma tupla com 2 semi-arestas
    
    Para acessar a semi-aresta em sentido padrão correspondente à aresta de
    índice k, pode ser utilizado G.edge(k)[0] ou G.half_edge((k, 0))
    Para acessar a semi-aresta em sentido contrário, pode ser utilizado
    G.edge(k)(1) ou G.half_edge((k, 1))
    """
    
    class Vertex(object):
        """
        Objeto representando um vértice da DCEL
        
        Tem as seguintes propriedades por padrão:
        - index: índice do vértice na lista de PlaneGraph G
        - name: nome do vértice, que pode ser definido livremente
        - incident_edge: uma semi-aresta que tem este vértice como origem
        - degree*: grau do vértice
        - coords: coordenadas do vértice, representado por uma tupla de floats
        
        *o grau do vértice é atualizado manualmente pelas operações do PlaneGraph
        para garantir tempo de consulta O(1)
        
        """
        
        def __init__(self, index, name, incident_edge=None, degree=0, coords=None):
            self._index = index
            self.name = name
            self.incident_edge = incident_edge
            self.degree = degree
            self.coords = coords

        def __eq__(self, other):
            return isinstance(other, PlaneGraph.Vertex) and self._index == other._index

        @property
        def _name(self):
            return "v" + self._index if self.name is None else self.name

        @property
        def index(self):
            return self._index

        @property
        def x(self):
            return self.coords[0]

        @property
        def y(self):
            return self.coords[1]

        def isAdjacent(self, other):
            first = self.incident_edge
            if first.destination == other: return True
            e = first.twin.next_edge
            while e != first:
                if e.destination == other: return True
                e = e.twin.next_edge
            return False

        def neighborhood(self):
            first = self.incident_edge
            vertices = [first.destination]
            e = first.twin.next_edge
            while e != first:
                vertices += [e.destination]
                e = e.twin.next_edge
            return vertices

        def incidentFaces(self):
            first = self.incident_edge
            faces = [first.incident_face]
            e = first.twin.next_edge
            while e != first:
                faces += [e.incident_face]
                e = e.twin.next_edge
            return faces

    class HalfEdge(object):
        """
        Objeto representando uma semi-aresta da DCEL
        
        Tem as seguintes propriedades por padrão:
        - index: índice da semi-aresta na lista de PlaneGraph G
        - name: nome da semi-aresta, que pode ser definido livremente
        - origin: vértice de origem da semi-aresta
        - twin*: semi-aresta gêmea
        - incident_face: face incidente à semi-aresta; como as semi-arestas devem
            estar em anti-horário no contorno de cada face, toda face incidente
            estará do lado esquerdo da semi-aresta
        - next_edge: próxima aresta no contorno da face
        - prev_edge: aresta anterior no contorno da face
        
        *se e é uma semi-aresta de u a v, a semi-aresta gêmea em relação a e é
        a semi-aresta de v a u
        
        Para acessar o destino de uma semi-aresta e, também pode ser utilizado
        e.destination
        
        A propriedade direction é utilizada pelo algoritmo de RectangularDrawing,
        e tanto a semi-aresta quanto sua gêmea recebem o mesmo valor
        
        A propriedade thickness é utilizada pelo algoritmo de RectangularGridDrawing,
        e tanto a semi-aresta quanto sua gêmea recebem o mesmo valor
        """
        
        def __init__(self, index, name, origin, twin=None, incident_face=None, nexte=None, preve=None):
            self._index = index
            self.name = name
            self.origin = origin
            self.twin = twin
            self.incident_face = incident_face
            self.next_edge = nexte
            self.prev_edge = preve
        
        def __eq__(self, other):
            return isinstance(other, PlaneGraph.HalfEdge) and self._index == other._index

        @property
        def _name(self):
            strid = str(self._index).replace("(","").replace(" ","").replace(")","")
            return "e" + strid if self.name is None else self.name

        @property
        def index(self):
            return self._index
        
        @property
        def destination(self):
            return self.twin.origin

        @property
        def direction(self):
            return self._direction

        @direction.setter
        def direction(self, direction):
            self._direction = direction
            self.twin._direction = direction

        @property
        def thickness(self):
            return self._thickness

        @thickness.setter
        def thickness(self, thickness):
            self._thickness = thickness
            self.twin._thickness = thickness
        
    class Face(object):
        """
        Objeto representando uma face da DCEL
        
        Tem as seguintes propriedades por padrão:
        - index: índice da face na lista de PlaneGraph G
        - name: nome da face, que pode ser definido livremente
        - length*: número de arestas no contorno da face
        - outer_component: uma semi-aresta do contorno da face, ou None se a
            face é externa
        - inner_component: uma semi-aresta do contorno da face, ou None se a
            face é interna
        
        *o número de arestas no contorno da face é atualizado manualmente pelas
        operações do PlaneGraph para garantir tempo de consulta O(1)
        
        """
        
        def __init__(self, index, length, name, outer_component=None, inner_component=None):
            self._index = index
            self.length = length
            self.name = name
            self.outer_component = outer_component
            self.inner_component = inner_component            
        
        def __len__(self):
            return self.length

        def __eq__(self, other):
            return isinstance(other, PlaneGraph.Face) and self._index == other._index

        @property
        def _name(self):
            return "f" + self._index if self.name is None else self.name

        @property
        def index(self):
            return self._index

        @property
        def isOuterFace(self):
            return self.index == 0

        @property
        def component(self):
            if self.isOuterFace: return self.inner_component
            return self.outer_component

        def boundaryCoords(self):
            first = self.component
            coords = [first.origin.coords]
            e = first.next_edge
            while e != first:
                coords += [e.origin.coords]
                e = e.next_edge
            return coords

    def __init__(self,
                 vertices=[(None, None), (None, None), (None, None)],
                 edges=[(None, (0, 1)), (None, (1, 2)), (None, (2, 0))],
                 faces=[(None, "+", [0, 1, 2]), (None, "-", [2, 1, 0])]):
        names = [v[0] for v in vertices]
        coords = [v[1] for v in vertices]
        _vertices = [self.Vertex(k, names[k], coords=coords[k]) for k in range(len(vertices))]

        _edges = []
        for k in range(len(edges)):
            name = edges[k][0]
            origin, destin = _vertices[edges[k][1][0]], _vertices[edges[k][1][1]]
            _edges += [(self.HalfEdge((k, 0), None if name is None else (name + "+"), origin),
                        self.HalfEdge((k, 1), None if name is None else (name + "-"), destin))]
            edge = _edges[-1]
            edge[0].twin = edge[1]
            edge[1].twin = edge[0]
            origin.incident_edge = edge[0]
            destin.incident_edge = edge[1]
            origin.degree += 1
            destin.degree += 1
        
        _faces = []
        for k in range(len(faces)):
            name = faces[k][0]
            orientation = faces[k][1]
            boundary = faces[k][2]
            first = _edges[boundary[0]][1 if orientation == "-" else 0]
            _faces += [self.Face(k, len(boundary), name, first)]
            face = _faces[-1]
            arcs = [first, None]
            for j in range(1, len(boundary)):
                arcs[1] = _edges[boundary[j]][0]
                if arcs[0].destination != arcs[1].origin: arcs[1] = arcs[1].twin
                arcs[1].incident_face = face
                arcs[0].next_edge = arcs[1]
                arcs[1].prev_edge = arcs[0]
                arcs[0] = arcs[1]
            first.incident_face = face
            arcs[0].next_edge = first
            first.prev_edge = arcs[0]
        _faces[0].inner_component = _faces[0].outer_component
        _faces[0].outer_component = None
        
        self._vertices = _vertices
        self._edges = _edges
        self._faces = _faces
    
    def __len__(self):
        return len(self._vertices)
    
    def __repr__(self):
        vertices = self._vertices
        edges = self._edges
        faces = self._faces

        text = "vertices=" + repr([(v.name, v.coords) for v in vertices])
        text += ", edges=" + repr([(e[0].name, (e[0].origin.index, e[0].destination.index)) for e in edges])
        lf = []
        for f in faces:
            first = f.component
            le = [first.index[0]]
            e = first.next_edge
            while e != first:
                le += [e.index[0]]
                e = e.next_edge
            lf += [(f.name, "+" if first.index[1] == 0 else "-", le)]
        text += ", faces=" + repr(lf)

        return text

    def __str__(self):
        vertices = self._vertices
        edges = self._edges
        faces = self._faces

        text = "{}\n".format(len(vertices))
        for v in vertices:
            text += "" if v.name is None else (v.name + " ")
            if v.coords is not None:
                text += "{:.4f} {:.4f}".format(v.x, v.y)
            text += "\n"

        text += "{}\n".format(len(edges))
        for e in edges:
            text += "" if e[0].name is None else (e[0].name + " ")
            text += "{} {}\n".format(e[0].origin.index, e[0].destination.index)

        text += "{}\n".format(len(faces))
        for f in faces:
            first = f.component
            text += "" if f.name is None else (f.name + " ")
            text += ("+" if first.index[1] == 0 else "-") + " {}".format(first.index[0])
            e = first.next_edge
            while e != first:
                text += " {}".format(e.index[0])
                e = e.next_edge
            text += "\n"

        return text

    @property
    def v(self):
        return len(self._vertices)

    @property
    def e(self):
        return len(self._edges)

    @property
    def f(self):
        return len(self._faces)

    def vertex(self, index):
        return self._vertices[index]
    
    def edge(self, index):
        return self._edges[index]
        
    def half_edge(self, index):
        return self._edges[index[0]][index[1]]
    
    def face(self, index):
        return self._faces[index]
    
    @property
    def vertices(self):
        return self._vertices

    @property
    def edges(self):
        return self._edges

    @property
    def faces(self):
        return self._faces

    @property
    def outer_face(self):
        return self._faces[0]

    @property
    def inner_faces(self):
        return self._faces[1:]

    @property
    def delta(self):
        return min([v.degree for v in self._vertices])

    @property
    def Delta(self):
        return max([v.degree for v in self._vertices])

    def renameVertex(self, index, name):
        self.vertex(index).name = name
    
    def renameEdge(self, index, name):
        self.edge(index)[0].name = name + "+"
        self.edge(index)[1].name = name + "-"
    
    def renameFace(self, index, name):
        self.face(index).name = name
    
    def setCoords(self, index, coords):
        self.vertex(index).coords = coords
    
    def unsetCoords(self, index):
        self.vertex(index).coords = None

    def clearCoords(self):
        for v in self.vertices: self.unsetCoords(v.index)

    def subdivide(self, e):
        """
        Subdivide a aresta e = (u, v) em duas arestas e devolve o novo vértice
        v' e a nova aresta e' resultante dessa operação
        
        Após a subdivisão, e = (u, v') e e' = (v', v)
        
        Complexidade: O(1)
        
        """
        k = self.v
        newv = self.Vertex(k, None, None)
        self._vertices += [newv]
        newv.degree = 2
        if not(None in e[0].origin.coords or None in e[0].destination.coords):
            self.setCoords(newv.index,
                           ((e[0].origin.x + e[0].destination.x)/2, (e[0].origin.y + e[0].destination.y)/2))
        k = self.e
        newe = (self.HalfEdge((k, 0), None, newv),
                self.HalfEdge((k, 1), None, e[0].destination))
        self._edges += [newe]
        newv.incident_edge = newe[0]
        newe[0].twin = newe[1]
        newe[1].twin = newe[0]
        newe[0].incident_face = e[0].incident_face
        newe[0].incident_face.length += 1
        newe[1].incident_face = e[1].incident_face
        newe[1].incident_face.length += 1
        newe[0].prev_edge = e[0]
        newe[0].next_edge = e[0].next_edge
        newe[1].prev_edge = e[1].prev_edge
        newe[1].next_edge = e[1]
        e[1].origin.incident_edge = newe[1]
        e[0].next_edge.prev_edge = newe[0]
        e[0].next_edge = newe[0]
        e[1].prev_edge.next_edge = newe[1]
        e[1].prev_edge = newe[1]
        e[1].origin = newv

        return newv, newe

    def join(self, v1, v2, f):
        """
        Liga o vértice v1 a v2 por dentro de f, criando uma nova aresta e' e
        uma nova face f'
        
        Após a operação, e' = (v1, v2), f terá em seu contorno a semi-aresta de
        e' com orientação padrão e f' terá em seu contorno a semi-aresta contrária;
        ou seja, a nova face f' surge à direita da semi-aresta padrão de e'
        
        Complexidade: O(|f'|)
        
        """
        if v1 == v2 or v1.isAdjacent(v2): return None, None
        k = self.e
        newe = (self.HalfEdge((k, 0), None, v1),
                self.HalfEdge((k, 1), None, v2))
        self._edges += [newe]
        newe[0].twin = newe[1]
        newe[1].twin = newe[0]
        k = self.f
        newf = self.Face(k, 1, None, newe[1])
        self._faces += [newf]
        newe[0].incident_face = f
        if f.isOuterFace: f.inner_component = newe[0]
        else: f.outer_component = newe[0]
        newe[1].incident_face = newf
        newf.outer_component = newe[1]
        e = v1.incident_edge
        while e.incident_face != f: e = e.twin.next_edge
        first = e
        first.incident_face = newf
        newf.length += 1
        while e.destination != v2:
            e = e.next_edge
            e.incident_face = newf
            newf.length += 1
        last = e
        f.length += 2 - newf.length
        newe[0].prev_edge = first.prev_edge
        newe[0].next_edge = last.next_edge
        newe[1].prev_edge = last
        newe[1].next_edge = first
        v1.incident_edge = newe[0]
        v1.degree += 1
        v2.incident_edge = newe[1]
        v2.degree += 1
        first.prev_edge.next_edge = newe[0]
        last.next_edge.prev_edge = newe[0]
        first.prev_edge = newe[1]
        last.next_edge = newe[1]

        return newe, newf

if __name__ == '__main__':
    vertices, edges, faces = intoargs(sys.stdin)

    G = PlaneGraph(vertices, edges, faces)

    if len(sys.argv) >= 2 and sys.argv[1] == "-r":
        print(repr(G))
    else:
        print(G)
