#!/usr/bin/python3

import os
import imageio
from math import sqrt

import matplotlib.pyplot as plt
import matplotlib.colors as clrs

def standardPlot(G, xlim=None, ylim=None, figsize=(4,4), title=None):
    """
    Plota o grafo plano G com os vértices desenhados como pontos e as arestas como
    segmentos de retas
    
    - xlim e ylim controlam os limites do eixos X e Y
    - figsize controla o tamanho da figura plotada
    - title pode ser utilizado para definir um título para o plot
    """
    _setfigure(figsize)
    _setlim(xlim, ylim)
    _plotedges(G, basecolor='black')
    _plotvertices(G, basecolor='black')
    if title is not None: plt.title(title)
    plt.show()

def indexedPlot(G, index=(True, True, True), xlim=None, ylim=None, figsize=(4,4), title=None):
    """
    Plota o grafo plano G com os índices de cada vértice, aresta e/ou face
    
    Sse index[0] = True, os índices dos vértices aparecerão no plot
    Sse index[1] = True, os índices das arestas aparecerão no plot
    Sse index[2] = True, os índices das faces aparecerão no plot
    
    - xlim e ylim controlam os limites do eixos X e Y
    - figsize controla o tamanho da figura plotada
    - title pode ser utilizado para definir um título para o plot
    """
    _setfigure(figsize)
    _setlim(xlim, ylim)
    _plotedges(G, index=index[1])
    _plotvertices(G, index=index[0])
    _plotfaces(G, index=index[2])
    if title is not None: plt.title(title)
    plt.show()

def labeledPlot(G, label=(True, True, True), xlim=None, ylim=None, figsize=(4,4), title=None):
    """
    Plota o grafo plano G destacando os elementos nomeados, ou seja, vértices, arestas
    ou faces que tenham nome diferente de None
    
    Sse label[0] = True, os vértices serão destacados
    Sse label[1] = True, as arestas serão destacadas
    Sse label[2] = True, as faces serão destacadas
    
    - xlim e ylim controlam os limites do eixos X e Y
    - figsize controla o tamanho da figura plotada
    - title pode ser utilizado para definir um título para o plot
    """
    _setfigure(figsize)
    _setlim(xlim, ylim)
    _plotedges(G, label=label[1], highlightcolor='tomato')
    _plotvertices(G, label=label[0], highlightcolor='seagreen')
    _plotfaces(G, label=label[2], highlightcolor='royalblue')
    if title is not None: plt.title(title)
    plt.show()

def highlightedSkeletonPlot(G, vertices=[], edges=[], orientation='+', xlim=None, ylim=None, figsize=(4,4), title=None):
    """
    Plota o grafo plano G destacando os vértices de índice i, com i na lista vertices,
    e as arestas de índice j, com j na lista edges
    
    Se orientation = '+': as arestas serão desenhadas como setas no sentido padrão
    Se orientation = '-': as arestas serão desenhadas como setas no sentido contrário ao padrão
    Se orientation = None: as arestas serão desenhadas como segmentos
    
    - xlim e ylim controlam os limites do eixos X e Y
    - figsize controla o tamanho da figura plotada
    - title pode ser utilizado para definir um título para o plot
    """
    _setfigure(figsize)
    _setlim(xlim, ylim)
    _plotedges(G, highlight=edges, orientation=orientation)
    _plotvertices(G, highlight=vertices)
    if title is not None: plt.title(title)
    plt.show()

def highlightedFacialPlot(G, faces=[], xlim=None, ylim=None, figsize=(4,4), title=None):
    """
    Plota o grafo plano G com as faces de índice k destacadas, em que k esteja 
    na lista faces 
    
    - xlim e ylim controlam os limites do eixos X e Y
    - figsize controla o tamanho da figura plotada
    - title pode ser utilizado para definir um título para o plot
    """
    _setfigure(figsize)
    _setlim(xlim, ylim)
    _plotedges(G)
    _plotvertices(G)
    _plotfaces(G, highlight=faces)
    if title is not None: plt.title(title)
    plt.show()

def _customizedPlot(G, vertex_texts, edge_texts, edge_colors, edge_styles, faces, ignoreText=False):
    plt.clf()
    _customizedPlotfaces(G, faces)
    _customizedPlotedges(G, edge_texts, edge_colors, edge_styles, ignoreText)
    _customizedPlotvertices(G, vertex_texts)

def _customizedPlotfaces(G, faces):
    for k in faces:
        f = G.face(k)
        coords = f.boundaryCoords()
        x = [coord[0] for coord in coords]
        y = [coord[1] for coord in coords]
        
        _pltface(x, y, color='turquoise', orientation=None, zorder=2)    

def _customizedPlotedges(G, edge_texts, edge_colors, edge_styles, ignoreText=False):
    for e in G.edges:
        style = edge_styles[e[0].index[0]]
        if style == '': continue
        x = [e[0].origin.x, e[0].destination.x]
        y = [e[0].origin.y, e[0].destination.y]
        level = edge_colors[e[0].index[0]]
        annotation = edge_texts[e[0].index[0]]
        color = ['silver', 'royalblue', 'tomato', 'seagreen'][level]
        if annotation in ['V', "E", "W"]:
            color = ['silver', 'royalblue', 'gold', 'violet'][level]
        zorder = (level+1) + 2*(level>0)
        width = 1.5
        if style == '=': width, style = 2.25, '-'
        if ignoreText: annotation = None
        
        plt.plot(x, y, linestyle=style, linewidth=width, marker='', color=color, zorder=zorder)
        plt.plot(x[0], y[0], marker='o', color=color, zorder=zorder)
        plt.plot(x[1], y[1], marker='o', color=color, zorder=zorder)
        plt.annotate(annotation, ((x[0] + x[1])/2, (y[0] + y[1])/2), zorder=7)

def _customizedPlotvertices(G, vertex_texts):
    for v in G.vertices:
        annotation = vertex_texts[v.index]
        plt.annotate(annotation, (v.x, v.y), zorder=8)

def _prepareForFrames(path, figsize=(4,4)):
    if not os.path.isdir(path): os.mkdir(path)
    _setfigure(figsize)

def _saveFrame(path, frame_counter):
    plt.savefig(os.path.join(path, str(frame_counter) + ".png"))

def _createAnimation(path, fps):
    frames = []
    files = [f.rsplit('.', 1) for f in os.listdir(path)]
    filtered_files = filter(lambda f: f[0].isdigit() and f[1] == "png", files)
    ordered_files = list(str(f) + ".png" for f in sorted(int(f[0]) for f in filtered_files))
    for file_name in ordered_files:
        frames.append(imageio.imread(os.path.join(path, file_name)))
    imageio.mimsave(path + ".gif", frames, fps=fps)

def _pltarrow(x, y, color, zorder):
    x0, y0 = x[0], y[0]
    dx, dy = x[1] - x0, y[1] - y0
    d = sqrt(dx**2 +dy**2)
    plt.arrow(x0, y0, dx, dy, color=color, head_width = 0.2 * d, head_length = 0.2 * d,
              length_includes_head=True, shape='left', zorder=zorder)

def _pltface(x, y, color, orientation, zorder):
    rgbacolor = clrs.to_rgba(color)
    transcolor = rgbacolor[:3] + (rgbacolor[3]/2,)
    plt.fill(x, y, facecolor=transcolor, edgecolor='none', zorder=zorder)
    for k in range(len(x)):
        x_ = [x[k-1], x[k]]
        y_ = [y[k-1], y[k]]
        if orientation == '+':
            _pltarrow(x_, y_, color=color, zorder=zorder+1)
        elif orientation == '-':
            _pltarrow(reversed(x_), reversed(y_), color=color, zorder=zorder+1)
        else:
            plt.plot(x, y, color=color, zorder=zorder+1)

def _setlim(xlim=None, ylim=None):
    if xlim is not None: plt.xlim(xlim)
    if ylim is not None: plt.ylim(ylim)

def _setfigure(figsize=(4,4)):
    plt.figure(figsize=figsize)

def _plotedges(G, index=False, label=False, highlight=[], orientation='+', basecolor='silver', highlightcolor='violet'):
    for e in G.edges:
        x = [e[0].origin.x, e[0].destination.x]
        y = [e[0].origin.y, e[0].destination.y]

        color = basecolor
        annotation = None
        if index:
            annotation = e[0].index[0]
        elif label and e[0].name is not None:
            color = highlightcolor
            annotation = e[0].name[:-1]

        plt.plot(x, y, marker='', color=color, zorder=1)
        if annotation is not None:
            plt.annotate(annotation, ((x[0] + x[1])/2, (y[0] + y[1])/2), zorder=9)

    for k in highlight:
        e = G.edge(k)
        j = (0 if orientation == '+' else 1)
        x = [e[j].origin.x, e[j].destination.x]
        y = [e[j].origin.y, e[j].destination.y]
        if orientation is None: plt.plot(x, y, color=highlightcolor, zorder=2)
        else: _pltarrow(x, y, color=highlightcolor, zorder=2)

def _plotvertices(G, index=False, label=False, highlight=[], basecolor='silver', highlightcolor='gold'):
    for v in G.vertices:
        color = basecolor
        annotation = None
        if index:
            annotation = v.index
        elif label and v.name is not None:
            color = highlightcolor
            annotation = v.name

        plt.plot(v.x, v.y, marker='o', color=color, zorder=3)
        plt.annotate(annotation, (v.x, v.y), zorder=9)

    for k in highlight:
        v = G.vertex(k)
        plt.plot(v.x, v.y, marker='o', color=highlightcolor, zorder=4)

def _plotfaces(G, index=False, label=False, highlight=[], highlightcolor='turquoise'):
    # index xor label necessarily
    if index or label:
        for f in G.faces:
            if label and f.name is None: continue

            annotation = f.index
            xytext = (0.05, 0.95)
            textcoords = 'axes fraction'

            if index and f.isOuterFace:
                plt.annotate(annotation, xytext, xytext=xytext,
                             textcoords=textcoords, zorder=9)
                continue

            coords = f.boundaryCoords()
            x = [coord[0] for coord in coords]
            y = [coord[1] for coord in coords]

            if not f.isOuterFace:
                xytext = (sum(x)/len(f), sum(y)/len(f))
                textcoords = 'data'
            # if label, f.name is not None
            if label:
                annotation = f.name
                _pltface(x, y, color=highlightcolor, orientation='+', zorder=5)

            plt.annotate(annotation, xytext, xytext=xytext,
                         textcoords=textcoords, zorder=9)

    for k in highlight:
        f = G.face(k)
        coords = f.boundaryCoords()
        x = [coord[0] for coord in coords]
        y = [coord[1] for coord in coords]
        _pltface(x, y, color=highlightcolor, orientation='+', zorder=7)
