#!/usr/bin/python3

import random
from math import log
from Utilities import unityRoot, segment, distance, distance2, normalizedVector, convex

def _positionOuterFace(G):
    listed_faces = [False] * G.f
    fixed_vertices = [False] * G.v
    faces = []

    f = G.outer_face
    e = f.inner_component
    n = len(f)
    coords = unityRoot(n)
    listed_faces[f.index] = True
    for k in range(n):
        v = e.origin
        G.setCoords(v.index, coords(n-k))
        fixed_vertices[v.index] = True
        if v.degree > 2:
            incfaces = v.incidentFaces()
            for g in incfaces:
                if not listed_faces[g.index]:
                    faces.append(g)
                    listed_faces[g.index] = True
        e = e.next_edge

    return listed_faces, fixed_vertices, faces

def _convexityHeuristic(e, fixed_vertices):
    v1 = e.origin
    lw, rw = e.prev_edge.twin, e.twin.next_edge
    while not fixed_vertices[lw.destination.index]: lw = lw.prev_edge.twin
    while not fixed_vertices[rw.destination.index]: rw = rw.twin.next_edge
    v3 = lw.destination
    v2 = rw.destination
    if convex(v1.coords, v2.coords, v3.coords):
        return ((v1.x + v2.x + v3.x)/3, (v1.y + v2.y + v3.y)/3)
    return v1.coords

# Algorithm for drawing 2-biconnected plane graphs based on heuristics 
def positionBiconnected(G, heuristic=True):
    listed_faces, fixed_vertices, faces = _positionOuterFace(G)

    # faces is a list of faces adjacent to some face in listed_faces
    while faces:
        f = faces.pop()
        e = f.outer_component
        while not fixed_vertices[e.origin.index]: e = e.prev_edge
        e0 = e.prev_edge
        # e0.destination is already fixed
        while e != e0:
            if fixed_vertices[e.destination.index]:
                e = e.next_edge
                continue
            # first.origin is fixed, .destination is not fixed
            first = e
            n = 0
            while not fixed_vertices[e.destination.index]:
                e = e.next_edge
                n += 1
            last = e
            # last.origin is not fixed, .destination is fixed
            # n = number of unfixed vertices between first and last
            if heuristic:
                p1 = _convexityHeuristic(first, fixed_vertices)
                p2 = _convexityHeuristic(last.twin, fixed_vertices)
            else:
                p1 = first.origin.coords
                p2 = last.destination.coords
            coords = segment(p1, p2)(n)

            e = first
            k = 1
            while not fixed_vertices[e.destination.index]:
                e = e.next_edge
                v = e.origin
                G.setCoords(v.index, coords(k))
                fixed_vertices[v.index] = True
                if v.degree > 2:
                    incfaces = v.incidentFaces()
                    for g in incfaces:
                        if not listed_faces[g.index]:
                            faces.append(g)
                            listed_faces[g.index] = True
                k += 1

def _EadesForce(G, u, c):
    force = [0, 0]
    for v in G.vertices:
        if u == v:
            continue
        elif u.isAdjacent(v):
            nv = normalizedVector(u.coords, v.coords)
            magnitude = c[0] * log(distance(u.coords, v.coords)/c[1])
        else:
            nv = normalizedVector(v.coords, u.coords)
            magnitude = c[2] / distance2(u.coords, v.coords)
        force[0] += magnitude * nv[0]
        force[1] += magnitude * nv[1]
    return force

def _faceCentroids(G):
    centroids = [(0,0)] * G.f
    for f in G.faces:
        coords = f.boundaryCoords()
        x = [coord[0] for coord in coords]
        y = [coord[1] for coord in coords]
        centroids[f.index] = (sum(x)/len(coords), sum(y)/len(coords))
    return centroids

def _adaptedForce(G, u, c, faces, centroids):
    force = _EadesForce(G, u, c)
    for f in G.inner_faces:
        coords = centroids[f.index]
        if f.index in faces:
            nv = normalizedVector(u.coords, coords)
            magnitude = c[0] * log(distance(u.coords, coords)/c[1])
        else:
            nv = normalizedVector(u.coords, coords)
            magnitude = c[2] / distance2(u.coords, coords)
        force[0] += c[4] * magnitude * nv[0]
        force[1] += c[4] * magnitude * nv[1]
    return force

# Eades's Force-Directed Algorithm [T16, Ead84] and an inspired/adapted version
#
# References:
# [T16]: Handbook of Graph Drawing and Visualization (TAMASSIA, R.)
# [Ead84]: A Heuristic for Graph Drawing (EADES, P.)
def positionSemifixed(G, adapted=False, xlim=[-1,1], ylim=[-1,1], seed=None, M=100, c=(2,1,1,0.1,0.05)):
    random.seed(seed)
    vertices = []
    for v in G.vertices:
        if v.coords is None:
            vertices += [v]
            G.setCoords(v.index, (random.uniform(xlim[0], xlim[1]), random.uniform(ylim[0], ylim[1])))
    if adapted:
        nonincfaces = [set(range(G.f)) - set([f.index for f in v.incidentFaces()]) for v in vertices]
    for _ in range(M):
        if adapted:
            centroids = _faceCentroids(G)
            k = 0
        forces = []
        for u in vertices:
            if adapted:
                forces += [_adaptedForce(G, u, c, nonincfaces[k], centroids)]
                k += 1
            else:
                forces += [_EadesForce(G, u, c)]
        k = 0
        for u in vertices:
            coords = (u.x + c[3] * forces[k][0], u.y + c[3] * forces[k][1])
            G.setCoords(u.index, coords)
            k += 1
