#!/usr/bin/python3

from NoCertificateConstructor import *
from Plotters import _prepareForFrames, _customizedPlot, _saveFrame, _createAnimation

# CONSTANTES
CARDINALS = ["N", "E", "S", "W"]
ORDINALS = ["NW", "NE", "SE", "SW"]
DIRECTIONS = ["horizontal", "vertical"]
NW, NE, SE, SW = 0, 1, 2, 3

# VARIÁVEIS GLOBAIS
DEBUG = False
MEMOIZED = True
ANIMATED = False
FRAME_COUNTER = 0
PATH = 'retangular'

def AnimatedDrawGraph(G, corners, path='retangular', fps=2, figsize=(4,4), memoized=True):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO, uma string path incidando um caminho para
    um diretório existente e um número indicando o FPS (frames por segundo)
    
    Devolve o resultado de DrawGraph(G, corners, memoized)
    
    Salva um .gif no caminho e nome indicado por path e as figuras dos frames
    gerados pelo desenho retangular de G no diretório path.
    
    Exemplo:
        path = '/home/user/output'
    Resultado:
        '/home/user/output.gif' e
        '/home/user/output/0.png', .., '/home/user/output/20.png'
    
    O parâmetro memoized controla se haverá memoização, ou seja, se serão
    mantidas as variáveis relativas ao caminho mais à esquerda P para uso em
    níveis mais profundos da recursão
    """
    global ANIMATED, PATH, FRAME_COUNTER
    hasCoords = True
    for v in G.vertices:
        if type(v.x) != int and type(v.x) != float: hasCoords = False
        if type(v.y) != int and type(v.y) != float: hasCoords = False
    if not hasCoords:
        raise Exception("G has vertices without coordinates!")
    ANIMATED = True
    PATH = path
    FRAME_COUNTER = 0
    _prepareForFrames(path, figsize)
    result = DrawGraph(G, corners, memoized)
    _createLastFrame(G, result[0], result[1])
    _createAnimation(path, fps)
    ANIMATED = False
    return result
    
def DrawGraph(G, corners, memoized=True):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO e devolve success, certificate,
    com success assumindo os valores True ou False
    
    Se G tem desenho retangular com os 4 corners fixos, success = True e 
    certificate é uma lista de direções para as arestas
    Caso contrário, success = False e certificate é uma lista de semi-arestas
    formando um ciclo ruim
    
    O parâmetro memoized controla se haverá memoização, ou seja, se serão
    mantidas as variáveis relativas ao caminho mais à esquerda P para uso em
    níveis mais profundos da recursão
    """
    global MEMOIZED
    MEMOIZED = memoized
    for e in G.edges:
        e[0].direction = None
        e[0].label, e[1].label = None, None
        e[0].is_on_p, e[1].is_on_p = False, False
    for f in G.faces: f.label = None
    prepareOuterFace(G, corners)
    success, certificate = Draw(G, corners)
    if not success: return success, certificate
    return True, [e[0].direction for e in G.edges]

def setDirection(e, direction, side, twin_side):
    """
    Recebe uma semi-aresta e, uma direção direction ("horizontal" ou "vertical") e
    dois rótulos side e twin_side ("N", "E", "S" ou "W")
    
    Atribui a direção para e e e.twin e também atribui os rótulos side e twin_side para
    e e e.twin, respectivamente
    """
    if e.direction is None:
        e.direction = direction
    elif e.direction != direction:
        raise Exception("Direction for G.edge({}) is already defined.".format(e.index[0]))
    e.label = side
    e.twin.label = twin_side

def isInnerEdge(G, e):
    """
    Devolve se a semi-aresta e é interna
    """
    return e.label is None

def isOuterEdge(G, e):
    """
    Devolve se a semi-aresta e é externa (faz parte de C_o)
    """
    return e.label is not None

def CardinalDistance(c1, c2):
    """
    Recebe c1, c2 cardinais (N, E, S ou W) e devolve (f(C1) - f(C2)) mod 4,
    sendo f(N) = 0, f(E) = 1, F(S) = 2 e f(W) = 3
    """
    return (CARDINALS.index(c1) - CARDINALS.index(c2)) % 4

def prepareOuterFace(G, corners):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO
    
    Atribui direções e rótulos para todas as arestas do contorno externo C_o(G)
    
    Devolve a semi-aresta de rótulo N que aponta para o canto NO
    """
    first = corners[0].incident_edge
    while first.incident_face != G.outer_face: first = first.twin.next_edge
    e = first
    k = 0
    setDirection(e, DIRECTIONS[k%2], CARDINALS[(k+2)%4], CARDINALS[k])
    e = e.next_edge
    while e != first:
        if e.origin in corners: k += 1
        setDirection(e, DIRECTIONS[k%2], CARDINALS[(k+2)%4], CARDINALS[k])
        e = e.next_edge
    return first.twin

def Draw(G, corners, update_side=None, memo=(None, None, None)):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, em ordem horária a partir de NO, uma variável update_side 
    contendo string ou None para indicar as bordas de H que devem ser atualizadas
    e uma tupla (a_list, b_list, p) para memoização
    
    Devolve success, certificate, com success assumindo os valores True ou False
    
    Se H tem desenho retangular com os 4 corners fixos, success = True e 
    certificate é None
    Caso contrário, success = False e certificate é uma lista de semi-arestas
    formando um ciclo ruim
    
    update_side pode assumir os valores "N", "E", "S", "W", "NS", "EW", "NESW" ou None
    
    No momento em que Draw é chamado para H, é garantido que as semi-arestas de
    C_o(H) tem direções e rótulos atribuídos e todas as semi-arestas do do interior
    de C_o(H) são arestas internas
    """
    if not MEMOIZED: memo = (None, None, None)
    if DEBUG: print("DRAW " + str([v.index for v in corners]))
    updateOuterEdges(G, corners, update_side)
    if ANIMATED: _createMidFrame(G, corners, memo[2], [])
    if isSquare(G, corners, update_side): return True, None
    success, certificate, separators = BoundariesUpdate(G, corners, update_side)
    if not success: return success, certificate
    if len(separators) > 0:
        if ANIMATED: _createMidFrame(G, corners, memo[2], separators)
        success, certificate, H_attrs  = separateComponents(G, corners, separators, memo)
        if not success: return success, certificate
        if ANIMATED: _createMidFrame(G, corners, memo[2], [])
        if DEBUG:
            for corners, _, _ in H_attrs: print(" H_ATTRS " + str([v.index for v in corners]))
        for corners, update_side, memo in H_attrs:
            success, certificate = Draw(G, corners, update_side, memo)
            if not success: return success, certificate
        return True, None
    else:
        if DEBUG: print(" WESTMOST")
        success, certificate, H_attrs = Partition(G, corners, memo)
        if not success: return success, certificate
        if ANIMATED: _createMidFrame(G, corners, H_attrs[0][2][2], [])
        if DEBUG:
            for corners, _, _ in H_attrs: print(" H_ATTRS " + str([v.index for v in corners]))
        for corners, update_side, memo in H_attrs:
            success, certificate = Draw(G, corners, update_side, memo)
            if not success: return success, certificate
        return True, None

def updateOuterEdges(G, corners, update_side=None):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando as bordas de H que devem ser atualizadas
    
    Atualiza os ponteiros e.outer_next e e.outer_prev de cada semi-aresta e
    do contorno externo C_o(H)
    
    Para cada e de C_o(H), e.outer_next é a próxima semi-aresta de C_o(H) em
    sentido horário e e.outer_prev é a semi-aresta anterior
    
    update_side pode assumir os valores "N", "E", "S", "W", "NS", "EW", "NESW" ou None 
    """
    if update_side is None: update_side = "NESW"
    for side in update_side: updateOuterSideEdges(G, corners, side)

def updateOuterSideEdges(G, corners, update_side=None):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando a borda de H que deve ser atualizada
    
    Atualiza os ponteiros e.outer_next e e.outer_prev de cada semi-aresta e
    do contorno externo C_o(H)
    
    update_side pode assumir os valores "N", "E", "S", "W" ou None
    """
    first = StartingHalfEdge(G, corners, update_side)

    first.next_edge.outer_next = first
    first.outer_prev = first.next_edge

    e1 = first
    corner = corners[(CARDINALS.index(update_side) + 1) % 4]

    e2 = e1.prev_edge
    if isInnerEdge(G, e2): e2 = e2.twin.prev_edge
    e1.outer_next = e2
    e2.outer_prev = e1
    e1 = e2
    while e1.destination != corner:
        e2 = e1.prev_edge
        if isInnerEdge(G, e2): e2 = e2.twin.prev_edge
        e1.outer_next = e2
        e2.outer_prev = e1
        e1 = e2

def StartingHalfEdge(G, corners, update_side=None):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando a borda de H que deve ser atualizada
    
    Devolve a semi-aresta e de C_o(H) com rótulo update_side apontando para um
    canto X, sendo:
    - X = NO, se update_side = N
    - X = NE, se update_side = E
    - X = SE, se update_side = S
    - X = SO, se update_side = O
    
    update_side pode assumir os valores "N", "E", "S", "W" ou None
    Se update_side é None, assume-se update_side = "N"
    """
    if update_side is None: update_side = "N"
    v = corners[CARDINALS.index(update_side)]
    if update_side in ["N", "S"]:
        first = v.incident_edge.twin
        e = first
        if e.label == update_side: return e
        e = e.next_edge.twin
        while e != first:
            if e.label == update_side: return e
            e = e.next_edge.twin
    elif update_side in ["E", "W"]:
        reference_side = "N" if update_side == "E" else "S"
        first = v.incident_edge
        e = first
        if e.label == reference_side: return e.prev_edge
        e = e.twin.next_edge
        while e != first:
            if e.label == reference_side: return e.prev_edge
            e = e.twin.next_edge
    raise Exception("Starting HalfEdge not found.")

def isSquare(G, corners, update_side=None):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando as bordas de H que devem ser atualizadas
    
    Devolve True se e somente se H é um ciclo
    
    update_side pode assumir os valores "N", "E", "S", "W", "NS", "EW", "NESW" ou None
    """
    if update_side is None: update_side = "N"
    first = StartingHalfEdge(G, corners, update_side[0])
    e = first.next_edge
    while e != first:
        if isInnerEdge(G, e): return False
        e = e.next_edge
    return True

def BoundariesUpdate(G, corners, update_side=None):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando as bordas de H que devem ter os rótulos atualizados
    
    Devolve success, certificate, separators, com success assumindo os valores
    True ou False e separators sendo uma lista de faces; e atualiza os rótulos
    das faces de acordo com os caminhos de borda (NN, NE, etc.)
    
    Se H não tem borda ou canto ruim, success = True e certificate = None
    Caso contrário, success = False e certificate é um ciclo ruim
    
    separators é uma lista contendo as faces separadoras de H em ordem de norte
    a sul ou oeste a leste
    
    update_side pode assumir os valores "N", "E", "S", "W", "NS", "EW", "NESW" ou None
    """
    if update_side is None or update_side == "NESW": update_side = [None]
    for side in update_side[::-1]:
        success, certificate, separators = BoundaryUpdate(G, corners, side)
        if not success: return success, certificate, None
    if update_side[0] == "W" or update_side[0] == "S":
        separators = list(reversed(separators))

    if update_side[0] == None: update_side = "NESW"
    for side in update_side[::-1]:
        success, certificate = checkBoundaryCycles(G, corners, side)
        if not success: return success, certificate, None

    return True, None, separators

def BoundaryUpdate(G, corners, update_side=None):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando a borda de H que deve ter os rótulos atualizados
    
    Devolve success, certificate, separators, com success assumindo os valores
    True ou False e separators sendo uma lista de faces; e atualiza os rótulos
    das faces de acordo com os caminhos de borda (NN, NE, etc.)
    
    Se nenhuma borda ou canto ruim for encontrada ao analisar os caminhos internos
    maximais das faces na borda determinada por update_side, success = True e
    certificate = None
    Caso contrário, success = False e certificate é um ciclo ruim
    
    separators é uma lista contendo as faces separadoras de H
    
    update_side pode assumir os valores "N", "E", "S", "W" ou None
    Se update_side = None, todas as bordas de H serão atualizadas
    """
    first = StartingHalfEdge(G, corners, update_side)
    separators = []
    
    e = first
    e.incident_face.label = None
    e = e.outer_next
    while e != first and (update_side is None or e.label == update_side):
        e.incident_face.label = None
        e = e.outer_next
    
    e = first
    success, certificate, separating = BoundaryFaceUpdate(G, corners, e.incident_face, e)
    if not success: return success, certificate, None
    if separating: separators += [e.incident_face]
    e = e.outer_next
    while e != first and (update_side is None or e.label == update_side):
        if e.incident_face.label is None:
            success, certificate, separating = BoundaryFaceUpdate(G, corners, e.incident_face, e)
            if not success: return success, certificate, None
            if separating: separators += [e.incident_face]
        e = e.outer_next
    
    return True, None, separators

def BoundaryFaceUpdate(G, corners, f, first):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, uma face f e
    uma aresta first sendo a primeira aresta (em sentido anti-horário) da borda
    de H no contorno de f
    
    Devolve success, certificate e separating, com success e separating assumindo
    os valores True ou False; e atualiza o rótulo da face f de acordo com o
    caminho de borda correspondente (NN, NE, etc.)
    
    Se nenhuma borda ou canto ruim for encontrada ao analisar os caminhos internos
    maximais de f, success = True e certificate = None
    Caso contrário, success = False e certificate é um ciclo ruim
    Se e somente se f é uma face separadora, separating = True
    
    O contorno de f é decomposto em subcaminhos internos e externos maximais
    Se H tem desenho retangular, f deve ter apenas um subcaminho interno maximal
    se não for separadora e f não pode ter um NW-, EN-, SE- ou WS-caminho de borda
    (se e somente se tiver, terá um canto ruim), considerando os rótulos em sentido
    anti-horário
    """
    dstring = ""

    e = first
    e = e.next_edge
    while e != first:
        if isInnerEdge(G, e):
            e, labels = passInnerPath(G, corners, e)
            dstring += str(CardinalDistance(labels[1], labels[0]))
            if dstring == "3": # NW-, EN-, SE- ou WS-caminho de borda
                return False, BadCycle(G, corners, e.next_edge, 1), False
            if len(dstring) > 1 and dstring != "22": # f deve ter apenas um subcaminho interno maximal
                return False, BadCycle(G, corners, e.next_edge, 2), False
            f.label = labels[0] + labels[1]
        e = e.next_edge

    d = int(dstring[0])
    return True, None, d == 2

def passInnerPath(G, corners, e):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e uma semi-aresta
    interna e tal que e.prev_edge é externa
    
    Devolve a última aresta do subcaminho interno maximal de e e o rótulo desse
    subcaminho
    """
    label1 = e.prev_edge.label
    while isInnerEdge(G, e.next_edge): e = e.next_edge
    label2 = e.next_edge.label
    return e, (label1, label2)

def checkBoundaryCycles(G, corners, update_side):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e update_side
    indicando a borda que deve ser verificada
    
    Devolve success, certificate, com success assumindo os valores True ou False
    
    Se nenhuma borda ruim for encontrada na borda verificada, devolve success = True
    e certificate = None
    Caso contrário, devolve success = False e certificate sendo uma lista de
    semi-arestas que formam o contorno da borda ruim
    
    As arestas do contorno de cada face de borda são numeradas de acordo com a
    ordem em que a face de borda aparece
    Se as arestas da borda não estiverem ordenadas, existe uma borda ruim com 
    n_{AH}(C) = 0
    Se duas faces que não são consecutivas compartilham aresta, então existe uma
    borda ruim com n_{AH}(C) = 1
    
    update_side pode assumir os valores "N", "E", "S" ou "W"
    Se update_side = None, todas as bordas de H serão verificadas
    """
    first = StartingHalfEdge(G, corners, update_side)

    e = first
    e.order = -1
    e = e.outer_next
    while e != first and e.label == update_side:
        e.order = -1
        e = e.outer_next

    e = first
    cleanFaceOrder(G, corners, e)
    e = e.outer_next
    while e != first and e.label == update_side:
        if e.order == -1: cleanFaceOrder(G, corners, e)
        e = e.outer_next

    e = first
    order = 0
    success, certificate = checkFaceOrder(G, corners, e, order)
    order += 1
    e = e.outer_next
    while e != first and e.label == update_side:
        if e.order is None:
            success, certificate = checkFaceOrder(G, corners, e, order)
            if not success: return success, certificate
            order += 1
        elif e.order != e.outer_prev.order:
            return False, BadCycle(G, corners, e, 0)
        e = e.outer_next

    return True, None

def cleanFaceOrder(G, corners, first):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H e first sendo uma
    semi-aresta
    
    Inicializa e.order e e.twin.order como sendo None, para todas as
    semi-arestas e do contorno da face incidente a first
    """
    e = first
    e.order = None
    e.twin.order = None
    e = e.next_edge
    while e != first:
        e.order = None
        e.twin.order = None
        e = e.next_edge

def checkFaceOrder(G, corners, first, order):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, first sendo uma
    semi-aresta e order um número inteiro
    
    Devolve success, certificate, com success assumindo os valores True ou False
    
    Se nenhuma perna de um ciclo de borda ruim com n_{AH}(C) = 1 for
    encontrada na borda verificada, success = True e certificate = None
    Caso contrário, success = False e certificate é uma lista de
    semi-arestas que formam o contorno da borda ruim
    
    As arestas do contorno da face de borda incidente a first que não estiverem
    numeradas recebem o número order
    Se uma aresta e já estiver numerada e a diferença em relação a order for
    maior que 1, e é perna anti-horária de uma borda ruim com n_{AH}(C) = 1
    """
    first.order = order
    first.twin.order = order
    e = first.next_edge
    while e != first:
        if e.order is None: e.twin.order = order
        e.order = order
        if e.order - e.twin.order > 1:
            return False, BadCycle(G, corners, e, 3)
        e = e.next_edge

    return True, None

def separateComponents(G, corners, separators, memo=(None, None, None)):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, separators sendo
    a lista ordenada das faces separadoras de H e uma tupla (a_list, b_list, p)
    para memoização
    
    As faces separadoras de H devem estar ordenadas de norte a sul ou oeste a leste
    
    Devolve success, certificate, H_attrs, com success assumindo os valores True ou False
    
    Se H não tiver nenhum ciclo interno ruim incidente a 2 faces separadoras,
    success = True e certificate = None
    Caso contrário, devolve success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim
    
    H_attrs é uma lista de tuplas (corners, update_side memo) identificando os subgrafos H
    que serão processados recursivamente
    """
    if separators[0].label in ["SN", "NS"]:
        return separateVertical(G, corners, separators, memo)
    elif separators[0].label in ["EW", "WE"]:
        return separateHorizontal(G, corners, separators, memo)
    raise Exception("Bad separators in separateComponents.")

def separateVertical(G, corners, separators, memo=(None, None, None)):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, separators sendo
    a lista das faces separadoras verticais de H e uma tupla (a_list, b_list, p)
    para memoização
    
    Devolve success, certificate, H_attrs, com success assumindo os valores True ou False
    
    Se H não tiver nenhum ciclo interno ruim incidente a 2 faces separadoras,
    success = True e certificate = None
    Caso contrário, devolve success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim
    
    H_attrs é uma lista de tuplas (corners, update_side memo) identificando os subgrafos H
    que serão processados recursivamente
    """
    
    _, _, p = memo
    if p: checkWesmostSeparatorLegs(G, corners, p, separators[0])
    
    new_corners_list = []
    for k in range(len(separators)):
        success, certificate, new_corners = separateVerticalFace(G, corners, separators, k)
        if not success: return success, certificate, None
        new_corners_list += [new_corners]

    corners_list = contractVerticalCornersList(new_corners_list)

    if DEBUG:
        print(" SEP_FACES " + str([s.index for s in separators]))
        for c in new_corners_list: print("  SEP_CORNERS " + str([v.index for v in c]))

    H_attrs = []
    if not (corners[NW] == corners_list[0][NW] and corners[SW] == corners_list[0][SW]):
        H_attrs += [((corners[NW], corners_list[0][NW], corners_list[0][SW], corners[SW]), "E", memo)]
    for j in range(1, len(corners_list)):
        H_attrs += [((corners_list[j-1][NE], corners_list[j][NW], corners_list[j][SW], corners_list[j-1][SE]), "EW", (None, None, None))]
    j = len(corners_list)-1
    if not (corners[NE] == corners_list[j][NE] and corners[SE] == corners_list[j][SE]):
        H_attrs += [((corners_list[j][NE], corners[NE], corners[SE], corners_list[j][SE]), "W", (None, None, None))]

    return True, None, H_attrs

def separateVerticalFace(G, corners, separators, k):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, separators sendo
    a lista das faces separadoras verticais de H e k um índice de separators
    
    Devolve success, certificate, corners, com success assumindo os valores True ou False
    
    Se H não tiver nenhum ciclo interno ruim incidente a borda leste da face
    separadora separators[k], success = True e certificate = None
    Caso contrário, devolve success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim
    Devolve corners como sendo os 4 cantos da face retangular separators[k]
    """
    corners = []
    e = separators[k].outer_component
    while e.label != "N": e = e.prev_edge
    while e.label == "N": e = e.next_edge
    e = e.prev_edge
    # e está apontando para o canto NW de separators[k]
    corners += [e.destination]
    while e.label == "N": e = e.prev_edge
    # e está apontando para o canto NE de separators[k]
    corners += [e.destination]
    if isInnerEdge(G, e):
        setDirection(e, "vertical", "E", "W")
        if k+1 < len(separators):
            isNeighbor = (e.twin.incident_face == separators[k+1])
        e = e.prev_edge
    while isInnerEdge(G, e):
        setDirection(e, "vertical", "E", "W")
        if k+1 < len(separators):
            if (e.twin.incident_face == separators[k+1]) != isNeighbor:
                return False, BadCycleBetween(G, e, separators[k], separators[k+1]), None
        e = e.prev_edge
    while e.label == "E": e = e.prev_edge
    # e está apontando para o canto SE de separators[k]
    corners += [e.destination]
    while e.label == "S": e = e.prev_edge
    # e está apontando para o canto SW de separators[k]
    corners += [e.destination]
    while isInnerEdge(G, e):
        setDirection(e, "vertical", "W", "E")
        e = e.prev_edge
    return True, None, corners

def contractVerticalCornersList(new_corners_list):
    """
    Recebe lista de cantos e contrai os cantos que compartilham borda leste/oeste
    """
    corners_list = [new_corners_list[0]]
    j = 0
    for k in range(1, len(new_corners_list)):
        corners1 = corners_list[j]
        corners2 = new_corners_list[k]
        if corners1[NE] == corners2[NW] and corners1[SE] == corners2[SW]:
            corners_list.pop()
            corners_list += [(corners1[NW], corners2[NE], corners2[SE], corners1[SW])]
        else:
            corners_list += [corners2]
            j += 1
    return corners_list

def separateHorizontal(G, corners, separators, memo=(None, None, None)):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, separators sendo
    a lista das faces separadoras horizontais de H e uma tupla (a_list, b_list, p)
    para memoização
    
    Devolve success, certificate, H_attrs, com success assumindo os valores True ou False
    
    Se H não tiver nenhum ciclo interno ruim incidente a 2 faces separadoras,
    success = True e certificate = None
    Caso contrário, devolve success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim
    
    H_attrs é uma lista de tuplas (corners, update_side memo) identificando os subgrafos H
    que serão processados recursivamente
    """
    new_corners_list = []
    for k in range(len(separators)):
        success, certificate, new_corners = separateHorizontalFace(G, corners, separators, k)
        if not success: return success, certificate, None
        new_corners_list += [new_corners]

    corners_list = contractHorizontalCornersList(new_corners_list)
    if DEBUG:
        print(" SEP_FACES " + str([s.index for s in separators]))
        for c in new_corners_list: print("  SEP_CORNERS " + str([v.index for v in c]))

    a_list, b_list, p = memo
    H_attrs = []
    if not (corners[NW] == corners_list[0][NW] and corners[NE] == corners_list[0][NE]):
        H_attrs += [((corners[NW], corners[NE], corners_list[0][NE], corners_list[0][NW]), "S", (a_list, None, p))]
    for j in range(1, len(corners_list)):
        H_attrs += [((corners_list[j-1][SW], corners_list[j-1][SE], corners_list[j][NE], corners_list[j][NW]), "NS", (None, None, p))]
    j = len(corners_list)-1
    if not (corners[SE] == corners_list[j][SE] and corners[SW] == corners_list[j][SW]):
        H_attrs += [((corners_list[j][SW], corners_list[j][SE], corners[SE], corners[SW]), "N", (None, b_list, p))]

    return True, None, H_attrs

def separateHorizontalFace(G, corners, separators, k):
    """
    Recebe um PlaneGraph G, os 4 cantos fixos de um subgrafo H, separators sendo
    a lista das faces separadoras horizontais de H e k um índice de separators
    
    Devolve success, certificate, corners, com success assumindo os valores True ou False
    
    Se H não tiver nenhum ciclo interno ruim incidente a borda sul da face
    separadora separators[k], success = True e certificate = None
    Caso contrário, devolve success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim
    Devolve corners como sendo os 4 cantos da face retangular separators[k]
    """
    corners = []
    e = separators[k].outer_component
    while e.label != "W": e = e.next_edge
    while e.label == "W": e = e.prev_edge
    # e está apontando para o canto NW de separators[k]
    corners += [e.destination]
    while isInnerEdge(G, e):
        setDirection(e, "horizontal", "N", "S")
        e = e.prev_edge
    while e.label == "N": e = e.prev_edge
    # e está apontando para o canto NE de separators[k]
    corners += [e.destination]
    while e.label == "E": e = e.prev_edge
    # e está apontando para o canto SE de separators[k]
    corners += [e.destination]
    if isInnerEdge(G, e):
        setDirection(e, "horizontal", "S", "N")
        if k+1 < len(separators):
            isNeighbor = (e.twin.incident_face == separators[k+1])
        e = e.prev_edge
    while isInnerEdge(G, e):
        setDirection(e, "horizontal", "S", "N")
        if k+1 < len(separators):
            if (e.twin.incident_face == separators[k+1]) != isNeighbor:
                return False, BadCycleBetween(G, e, separators[k], separators[k+1]), None
        e = e.prev_edge
    while e.label == "S": e = e.prev_edge
    # e está apontando para o canto SW de separators[k]
    corners += [e.destination]
    return True, None, corners

def contractHorizontalCornersList(new_corners_list):
    """
    Recebe lista de cantos e contrai os cantos que compartilham borda sul/norte
    """
    corners_list = [new_corners_list[0]]
    j = 0
    for k in range(1, len(new_corners_list)):
        corners1 = corners_list[j]
        corners2 = new_corners_list[k]
        if corners1[SE] == corners2[NE] and corners1[SW] == corners2[NW]:
            corners_list.pop()
            corners_list += [(corners1[NW], corners1[NE], corners2[SE], corners2[SW])]
        else:
            corners_list += [corners2]
            j += 1
    return corners_list

def Partition(G, corners, memo=(None, None, None)):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma tupla (a_list, b_list, p) para memoização
    
    Devolve success, certificate, H_attrs, com success assumindo os valores True ou False
    
    Se existir um ciclo interno ruim C' com n_AH(C) = 0 e n_H(C) = 1 ou
    existir um ciclo crítico interno ruim anexado a Q_H(C) de algum ciclo crítico
    maximal C do caso 3.2 devolve-se success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim C'
    Caso contrário, success = True e certificate = None
    
    H_attrs é uma lista de tuplas (corners, update_side memo) identificando os subgrafos H
    que serão processados recursivamente
    
    - P contem a lista de arestas que formam o NS-caminho mais à esquerda de H
    - a_list mantém uma lista de arestas que pertencem a um NN- ou NL-caminho de borda,
    ordenadas de acordo com a ordem das semi-arestas de p
    - b_list mantém uma lista de arestas que pertencem a um SS- ou LS-caminho de borda.
    ordenadas inversamente à ordem das semi-arestas de p

    Tanto P quanto a_list e b_list podem ser recebidos como listas vazias, sendo 
    necessário construí-los nesses casos. Jã a_list e b_list, especificamente, devem 
    ser atualizados em toda chamada de Partition
    """
    a_list, b_list, p = memo
    if not p:
        p = WestmostPath(G, corners)
        prepareClockwiseFaces(G, corners, p)
        countEdgeTraversals(G, corners, p)
        findCriticalCycles(G, corners, p, 0, len(p))
        countCCLegs(G, corners, p, 0, len(p))
    if ANIMATED: _createMidFrame(G, corners, p, [])
    a_list, b_list = updateEdgesMemo(G, corners, (a_list, b_list, p))
    if DEBUG:
        print("  P " + str([e.index for e in p]))
        print("  A " + str([e.index for e in a_list]))
        print("  B " + str([e.index for e in b_list]))
    leg = checkStartingPathLegs(G, corners, a_list[-1])
    if leg: correctCriticalCycles(G, corners, p, leg)
    leg = checkEndingPathLegs(G, corners, b_list[-1])
    if leg: correctCriticalCycles(G, corners, p, leg)
    return PartitionWithMemo(G, corners, (a_list, b_list, p))

def WestmostPath(G, corners):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H
    
    Devolve uma lista de semi-arestas que formam o NS-caminho mais à esquerda de H
    """
    p = []
    outer_edge = StartingHalfEdge(G, corners, "N")
    while outer_edge.origin.degree != 3: outer_edge = outer_edge.prev_edge
    first = outer_edge.prev_edge
    while outer_edge.label != "S":
        sub_p, outer_edge = WestmostSubpath(G, corners, first)
        p += sub_p
        first = p[-1].next_edge.twin

    # atribui índices às arestas de p
    for k in range(len(p)):
        p[k].p_index = k
        p[k].twin.p_index = k

    return p

def WestmostSubpath(G, corners, first):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma semi-aresta first
    
    Devolve o subcaminho do NS-caminho mais à esquerda P contido no contorno
    da face incidente a first
    """
    e = first
    sub_p, p = [], []
    while isInnerEdge(G, e):
        if e.destination.degree == 3:
            p += sub_p
            sub_p = []
        sub_p += [e.twin]
        e = e.prev_edge
    if e.label == "S": p += sub_p
    return p, e

def setCounter(G, e, count):
    """
    Recebe um PlaneGraph G, uma semi-aresta e e um número count
    
    Atribui o mesmo valor count ao counter de e e ao counter da semi-aresta gêmea
    de e
    """
    e.counter = count
    e.twin.counter = count

def prepareClockwiseFaces(G, corners, p):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e o NS-caminho mais à esquerda p
    
    Inicializa os contornos das faces anexadas a p com o o valor 0, marca cada
    semi-aresta de p e inicializa os vértices de p para a identificação de ciclos
    crtíticos maximais
    """
    for e in p:
        e.incident_face.prepared = False
    for e in p:
        if not e.incident_face.prepared:
            e.incident_face.prepared = True
            first = e
            setCounter(G, e, 0)
            e = e.next_edge
            while e != first:
                setCounter(G, e, 0)
                e = e.next_edge
    for e in p:
        e.is_on_p = True
        e.origin.end_type = None
    p[-1].destination.end_type = None

def countEdgeTraversals(G, corners, p):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e o NS-caminho mais à esquerda p
    
    Percorre o contorno de cada face anexada a P contando quantas vezes cada
    aresta foi percorrida
    """
    k = 0
    while k < len(p):
        e = p[k]
        while e.is_on_p:
            k += 1
            e = e.next_edge
        while e.origin.degree != 3: e = e.next_edge
        e = e.next_edge
        while not e.is_on_p:
            setCounter(G, e, e.counter+1)
            e = e.next_edge

def findCriticalCycles(G, corners, p, start, stop):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, o NS-caminho mais à esquerda p, e dois índices start e stop
    
    Detecta cada ciclo crítico maximal anexado ao subcaminho de p entre os índices
    start e stop
    
    Os vértices anteriores e posteriores de cada ciclo crítico maximal C recebe as
    marcas "H" e "T", respectivamente
    A perna de C incidente ao vértice anterior de C mantém referência à perna
    horária e à perna incidente ao vértice posterior de P
    Similarmente, a perna de C incidente ao vértice posterior de C mantém
    referência à perna horária e à perna incidente ao vértice anterior de P
    """
    k = start
    while k < stop:
        e = p[k]
        while e.is_on_p:
            k += 1
            e = e.next_edge
        neighbor_face = e.twin.incident_face
        head_edge = e.prev_edge
        found = False
        leg = None
        while not e.is_on_p:
            if e.counter == 2:
                if leg is None or e.twin.incident_face != leg.twin.incident_face:
                    if e.twin.incident_face != neighbor_face:
                        leg = e
                setCounter(G, e, -1)
            e = e.next_edge
        if e.p_index > head_edge.p_index:
            found = True
            leg = None
            tail_edge = e
        elif leg:
            found = True
            e = leg.twin
            while not e.is_on_p:
                if e.counter == 2: setCounter(G, e, -1)
                e = e.next_edge
            tail_edge = e
        if found:
            head_vertex, tail_vertex = head_edge.destination, tail_edge.origin
            head_vertex.end_type, tail_vertex.end_type = "H", "T"
            head_vertex.other_end, tail_vertex.other_end = tail_edge, head_edge
            if leg: head_vertex.leg, tail_vertex.leg = leg, leg.twin
            else: head_vertex.leg, tail_vertex.leg = None, None

def countCCLegs(G, corners, p, start, stop):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, o NS-caminho mais à esquerda p, e dois índices start e stop
    
    Conta o número de pernas anti-horárias de C para cada ciclo crítico maximal
    entre start e stop e armazena o valor nos vértices anteriores e posteriores
    """
    count = 0
    count_list = [0]
    for k in range(start+1, stop):
        if p[k].origin.degree == 3 and p[k].prev_edge.is_on_p: count += 1
        count_list += [count]
    for k in range(start+1, stop):
        if p[k].origin.end_type == "H":
            j = p[k].origin.other_end.p_index
            n = count_list[j - start] - count_list[k - start]
            head_vertex = p[k].origin
            tail_vertex = p[j].origin
            head_vertex.ncc, tail_vertex.ncc = n, n

def updateEdgesMemo(G, corners, memo):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma tupla (a_list, b_list, p) para memoização
    
    Devolve a_list e b_list atualizados para o subgrafo H atual
    """
    a_list, b_list, p = memo
    if a_list:
        while isOuterEdge(G, a_list[-1]): a_list.pop()
        first_index = a_list[-1].p_index
        last_index = LastPIndex(G, corners)
        b_list = EndingPathCandidates(G, corners, (a_list, b_list, p), (first_index, last_index))
    elif b_list:
        while isOuterEdge(G, b_list[-1]): b_list.pop()
        first_index = FirstPIndex(G, corners)
        last_index = b_list[-1].p_index
        a_list = StartingPathCandidates(G, corners, (a_list, b_list, p), (first_index, last_index))
    else:
        first_index = FirstPIndex(G, corners)
        last_index = LastPIndex(G, corners)
        a_list = StartingPathCandidates(G, corners, (a_list, b_list, p), (first_index, last_index))
        first_index = a_list[-1].p_index
        b_list = EndingPathCandidates(G, corners, (a_list, b_list, p), (first_index, last_index))

    return a_list, b_list

def FirstPIndex(G, corners):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H
    
    Devolve o índice da primeira semi-aresta de P', sendo P' o subcaminho
    do NS-caminho mais à esquerda P que está contido em H
    """
    e = StartingHalfEdge(G, corners, "N")
    while isOuterEdge(G, e): e = e.prev_edge
    return e.twin.p_index

def LastPIndex(G, corners):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H
    
    Devolve o índice da última semi-aresta de P', sendo P' o subcaminho 
    do NS-caminho mais à esquerda P que está contido em H
    """
    e = StartingHalfEdge(G, corners, "W")
    while isOuterEdge(G, e): e = e.next_edge
    return e.twin.p_index

def StartingPathCandidates(G, corners, memo, indices):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, uma tupla (a_list, b_list, p) para memoização e uma tupla
    de índices (first_index, last_index)
    
    Atualiza a_list adicionando semi-arestas que estejam contidas num NN- ou
    NL-caminho de borda de H e que estejam no subcaminho de p entre first_index e
    last_index
    """
    a_list, _, p = memo
    first_index, last_index = indices
    k = first_index
    if not a_list: a_list = [p[k]]
    if p[k].destination.end_type == "H": k = p[k].destination.other_end.p_index
    else: k += 1
    while k < last_index:
        if p[k].incident_face.label in ["NN", "NE", "EN"]: a_list += [p[k]]
        if p[k].destination.end_type == "H": k = p[k].destination.other_end.p_index
        else: k += 1
    return a_list

def EndingPathCandidates(G, corners, memo, indices):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, uma tupla (a_list, b_list, p) para memoização e uma tupla
    de índices (first_index, last_index)
    
    Atualiza b_list adicionando semi-arestas que estejam contidas num SS- ou
    LS-caminho de borda de H e que estejam no subcaminho de p entre first_index e
    last_index
    """
    _, b_list, p = memo
    first_index, last_index = indices
    k = last_index
    if not b_list: b_list = [p[k]]
    if p[k].origin.end_type == "T": k = p[k].origin.other_end.p_index
    else: k -= 1
    while k > first_index:
        if p[k].incident_face.label in ["SS", "ES", "ES"]: b_list += [p[k]]
        if p[k].origin.end_type == "T": k = p[k].origin.other_end.p_index
        else: k -= 1
    return b_list

def checkStartingPathLegs(G, corners, first):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma semi-aresta first
    
    Considere f como sendo a face incidente a first, então first é a última
    semi-aresta de f contida no NS-caminho mais à esquerda P, considerando a ordem
    das semi-arestas de P
    
    Verifica se existe um ciclo crítico maximal C com 1 perna horária e anexado a
    P tal que a face incidente a first contenha e e o vértice posterior de C.
    Caso exista tal C, então verifica se existe um ciclo crítico C' anexado a P
    com vértice anterior coincidente ao de C. Caso exista tal C', devolve a
    perna horária do C' maximal
    Caso contrário, devolve None
    """
    e = first
    leg = None
    while isInnerEdge(G, e):
        if e.is_on_p and e.origin.end_type == "T":
            if e.origin.leg:
                leg = e.origin.leg
                break
        e = e.prev_edge
    if not leg: return None
    last = e.origin.leg.twin
    e = e.origin.other_end.next_edge
    neighbor_face = e.twin.incident_face
    leg = None
    while e != last:
        if e.counter == -1:
            if leg is None or e.twin.incident_face != leg.twin.incident_face:
                if e.twin.incident_face != neighbor_face:
                    leg = e
        e = e.next_edge
    return leg

def checkEndingPathLegs(G, corners, first):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma semi-aresta first
    
    Considere f como sendo a face incidente a first, então first é a primeira
    semi-aresta de f contida no NS-caminho mais à esquerda P, considerando a ordem
    das semi-arestas de P

    Verifica se existe um ciclo crítico maximal C com 1 perna horária e anexado a
    P tal que a face incidente a first contenha e e o vértice anterior de C.
    Caso exista tal C, então verifica se existe um ciclo crítico C' anexado a P
    com vértice posterior coincidente ao de C. Caso exista tal C', devolve a
    perna horária do C' maximal
    Caso contrário, devolve None
    """
    e = first
    leg = None
    while isInnerEdge(G, e):
        if e.is_on_p and e.destination.end_type == "H":
            if e.destination.leg:
                leg = e.destination.leg
                break
        e = e.next_edge
    if not leg: return None
    last = e.destination.leg.twin
    e = e.destination.other_end.prev_edge
    neighbor_face = e.twin.incident_face
    leg = None
    while e != last:
        if e.counter == -1:
            if leg is None or e.twin.incident_face != leg.twin.incident_face:
                if e.twin.incident_face != neighbor_face:
                    leg = e
        e = e.prev_edge
    if not leg: return None
    return leg.twin

def checkWesmostSeparatorLegs(G, corners, p, f):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, o NS-caminho mais à esquerda p e uma face separadora
    vertical f
    
    Verifica se existem ciclos críticos maximais C anexados a P satisfazendo:
    C tem 1 perna horária e contida no contorno de f e existe um ciclo crítico
    C' anexado a P com vértice anterior ou posterior coincidente ao de C.
    
    Corrige a memoização caso exista tal C, pois o C' maximal será um ciclo
    crítico maximal ignorado por p
    """
    e = f.outer_component
    while e.label != "N": e = e.prev_edge
    while e.label == "N": e = e.next_edge
    # e está saindo do canto NW da face f
    while isInnerEdge(G, e):
        if e.is_on_p:
            leg = checkEndingPathLegs(G, corners, e)
            if leg:
                correctCriticalCycles(G, corners, p, leg)
            break
        e = e.next_edge
    while isInnerEdge(G, e): e = e.next_edge
    e = e.prev_edge
    # e está apontando para o canto SW da face f
    while isInnerEdge(G, e):
        if e.is_on_p:
            leg = checkStartingPathLegs(G, corners, e)
            if leg:
                correctCriticalCycles(G, corners, p, leg)
            break
        e = e.prev_edge

def correctCriticalCycles(G, corners, p, leg):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, o NS-caminho mais à esquerda p e a perna horária de um
    ciclo crítico maximal que não está incluído na memoização
    
    Corrige a memoização dos ciclos críticos maximais de p considerando C'
    como um dos ciclos críticos maximais
    """
    e = leg
    while not e.is_on_p: e = e.prev_edge
    head_edge = e
    e = leg.twin
    while not e.is_on_p: e = e.next_edge
    tail_edge = e
    head_vertex, tail_vertex = head_edge.destination, tail_edge.origin
    if head_vertex.end_type: head_vertex.other_end.origin.end_type = None
    if tail_vertex.end_type: tail_vertex.other_end.destination.end_type = None
    head_vertex.end_type, tail_vertex.end_type = "H", "T"
    head_vertex.other_end, tail_vertex.other_end = tail_edge, head_edge
    head_vertex.leg, tail_vertex.leg = leg, leg.twin
    start, stop = head_edge.p_index+1, tail_edge.p_index
    reprepareClockwiseFaces(G, corners, p, start, stop)
    e = leg
    while not e.is_on_p:
        if e.counter == 2: setCounter(G, e, -1)
        e = e.prev_edge
    e = leg.twin
    while not e.is_on_p:
        if e.counter == 2: setCounter(G, e, -1)
        e = e.next_edge
    findCriticalCycles(G, corners, p, start, stop)
    countCCLegs(G, corners, p, start-1, stop+1)

def reprepareClockwiseFaces(G, corners, p, start, stop):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, o NS-caminho mais à esquerda p, e dois índices start e stop
    
    Retoma a contagem das arestas das faces anexadas ao subcaminho de p entre os
    índices start e stop
    """
    for k in range(start, stop):
        p[k].incident_face.prepared = False
    for k in range(start, stop):
        if not p[k].incident_face.prepared:
            e = p[k]
            e.incident_face.prepared = True
            first = e
            if e.counter == -1: setCounter(G, e, 2)
            e = e.next_edge
            while e != first:
                if e.counter == -1: setCounter(G, e, 2)
                e = e.next_edge
    for k in range(start+1, stop):
        p[k].origin.end_type = None

def PartitionWithMemo(G, corners, memo):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e uma tupla (a_list, b_list, p) onde:
    - a_list é uma lista de semi-arestas de p que pertencem a NN- ou NL-caminhos 
    de borda, ordenadas de acordo com a ordem das semi-arestas de p
    - b_list é uma lista de semi-arestas de p que pertencem a SS- ou LS-caminhos
    de borda, ordenadas inversamente à ordem das semi-arestas de p
    - p é o NS-caminho mais à esquerda de H
    
    Devolve success, certificate, H_attrs, com success assumindo os valores True ou False
    
    Se existir um ciclo interno ruim C' com n_AH(C) = 0 e n_H(C) = 1 ou
    existir um ciclo crítico interno ruim anexado a Q_H(C) de algum ciclo crítico
    maximal C do caso 3.2 devolve-se success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim C'
    Caso contrário, success = True e certificate = None
    
    H_attrs é uma lista de tuplas (corners, update_side, memo) identificando os subgrafos H
    que serão processados recursivamente
    """
    a_list, b_list, p = memo
    north_corner = StartingPartition(G, corners, a_list)

    H_attrs = []
    k = a_list[-1].p_index
    while k <= b_list[-1].p_index:
        setDirection(p[k], "vertical", "W", "E")
        if p[k].destination.end_type == "H" and p[k].destination.other_end.p_index <= b_list[-1].p_index:
            head_edge = p[k]
            success, certificate, new_corners = partitionByCase(G, corners, memo, head_edge)
            if not success: return success, certificate, None
            if new_corners: H_attrs += [(new_corners, None, (None, None, None))]
            k = head_edge.destination.other_end.p_index
        else:
            k += 1

    south_corner = EndingPartition(G, corners, b_list)

    west_H_attr = [((corners[NW], north_corner, south_corner, corners[SW]), "E", memo)]
    east_H_attr = [((north_corner, corners[NE], corners[SE], south_corner), "W", (None, None, None))]

    return True, None, west_H_attr + H_attrs + east_H_attr

def StartingPartition(G, corners, a_list):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e a_list sendo a lista de semi-arestas de p que pertencem a
    NN- ou NL-caminhos de borda, ordenadas de acordo com a ordem das semi-arestas de p
    
    Utiliza a_list para particionar H por P_{ini} e devolve o vértice inicial de 
    P_{ini}
    """
    e = a_list[-1].prev_edge
    while isInnerEdge(G, e):
        setDirection(e, "vertical", "W", "E")
        e = e.prev_edge
    return e.destination

def EndingPartition(G, corners, b_list):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e b_list sendo a lista de semi-arestas de p que pertencem a
    SS- ou LS-caminhos de borda, ordenadas inversamente à ordem das semi-arestas de p
    
    Utiliza b_list para particionar H por P_{fim} e devolve o vértice final de 
    P_{fim}
    """
    e = b_list[-1].next_edge
    while isInnerEdge(G, e):
        setDirection(e, "vertical", "W", "E")
        e = e.next_edge
    return e.origin

def partitionByCase(G, corners, memo, head_edge):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, a tupla (a_list, b_list, p) e a perna de um ciclo crítico
    maximal C anexado ao NS-caminho mais à esquerda P incidente ao seu vértice
    anterior
    
    Devolve success, certificate, new_corners, com success assumindo os valores True ou False
    
    Se existir um ciclo interno ruim C' com n_AH(C) = 0 e n_H(C) = 1 ou
    existir um ciclo crítico interno ruim anexado a Q_H(C) de algum ciclo crítico
    maximal C do caso 3.2 devolve-se success = False e certificate é uma lista de
    semi-arestas que formam o contorno do ciclo ruim C'
    Caso contrário, success = True e certificate = None
    
    Se C cair em um caso em que P_H e P_{AH} formam um ciclo H, new_corners será
    os cantos de H
    Caso contrário, new_corners = None
    
    Particiona C de acordo com cada caso da definição de P_H e P_{AH}
    """
    leg = head_edge.destination.leg
    ncc = head_edge.destination.ncc
    new_corners = None
    if not leg:
        partitionClockwisePath(G, corners, memo, head_edge)
    elif ncc < 1:
        return False, BadBigCycle(G, leg, head_edge), None
    elif ncc == 1:
        new_corners = partitionBigCycle(G, corners, memo, head_edge)
    else:
        success, certificate, found, cycle_legs = findMaximalCriticalCycle(G, corners, memo, head_edge)
        if not success: return success, certificate, None
        if found:
            new_corners = partitionSmallCycle(G, corners, memo, head_edge, cycle_legs)
        else:
            partitionClockwisePath(G, corners, memo, head_edge)
    return True, None, new_corners

def partitionClockwisePath(G, corners, memo, head_edge):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, a tupla (a_list, b_list, p) e a perna de um ciclo crítico
    maximal C anexado ao NS-caminho mais à esquerda P incidente ao seu vértice
    anterior
    
    Particiona C de acordo com o caso em que n_{H}(C) = 0 ou n_{H}(C) = 1,
    n_{AH}(C) > 1 e não tem ciclo crítico anti-horário C' anexado a Q_H(C)
    """
    tail_edge = head_edge.destination.other_end
    leg = head_edge.destination.leg
    e = head_edge.next_edge
    if leg:
        while e != leg:
            setDirection(e, "vertical", "W", "E")
            e = e.next_edge
        e = e.twin.next_edge
    while e != tail_edge:
        setDirection(e, "vertical", "W", "E")
        e = e.next_edge

def partitionBigCycle(G, corners, memo, head_edge):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, a tupla (a_list, b_list, p) e a perna de um ciclo crítico
    maximal C anexado ao NS-caminho mais à esquerda P incidente ao seu vértice
    anterior
    
    Particiona C de acordo com o caso em que n_{H}(C) = 1 e n_{AH}(C) = 1
    
    Devolve os cantos do ciclo H que deve ser processado recursivamente
    """
    _, _, p = memo
    new_corners = [head_edge.destination]
    tail_edge = head_edge.destination.other_end
    leg = head_edge.destination.leg
    # lado horário
    e = head_edge.next_edge
    while e != leg:
        setDirection(e, "horizontal", "W", "N")
        e = e.next_edge
    new_corners += [e.origin]
    e = e.twin.next_edge
    while e != tail_edge:
        setDirection(e, "vertical", "W", "E")
        e = e.next_edge
    new_corners += [e.origin]
    # lado anti-horário
    k = head_edge.p_index + 1
    while not (p[k].origin.degree == 3 and p[k].prev_edge == p[k-1]):
        e = p[k]
        setDirection(e, "vertical", "W", "E")
        k += 1
    new_corners += [p[k].origin]
    while p[k] != tail_edge:
        e = p[k]
        setDirection(e, "horizontal", "S", "E")
        k += 1
    return tuple(new_corners)

def findMaximalCriticalCycle(G, corners, memo, head_edge):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, a tupla (a_list, b_list, p) e a perna de um ciclo crítico
    maximal C anexado ao NS-caminho mais à esquerda P incidente ao seu vértice
    anterior
    
    Devolve success, certificate, found e cycle_legs
    
    Particiona C de acordo com o caso em que n_{H}(C) = 1 e n_{AH}(C) > 1    
    
    Se G(C) tem um ciclo crítico anti-horário C' anexado a Q_H(C), found = True
    Caso contrário, found = False
    
    Quando found = True, se n_{AH}(C') = 0 ou n_{H}(C') = 0, devolve-se success = False e certificate
    é uma lista de semi-arestas que formam o contorno do ciclo C'
    Caso contrário, success = True e certificate = None
    
    Quando success = True e found = True, cycle_legs são as pernas de C' contidas
    em G(C)
    """
    ip = prepareInsidePath(G, corners, memo, head_edge)
    prepareInsideFacesOrder(G, corners, ip)

    order = 0
    leg_value, leg = findMaximumLeg(G, corners, ip[0], order)
    order += 1
    for k in range(1, len(ip)):
        if ip[k].order is None:
            value, leg_candidate = findMaximumLeg(G, corners, ip[k], order)
            if value > leg_value:
                leg = leg_candidate
                leg_value = value
            order += 1
        elif ip[k].order != ip[k-1].order:
            return False, BadCycleInside(G, None, head_edge.destination.leg, ip[k]), True, None
    if leg_value < 2:
        return True, None, False, None
    else:
        return findThreeLegs(G, corners, leg, head_edge.destination.leg)

def prepareInsidePath(G, corners, memo, head_edge):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, a tupla (a_list, b_list, p) e a perna de um ciclo crítico
    maximal C anexado ao NS-caminho mais à esquerda P incidente ao seu vértice
    anterior
    
    Sendo C um ciclo com n_{H}(C) = 1, devolve a lista de semi-arestas contidas
    em G(C) que formam o caminho Q_H(C)
    """
    tail_edge = head_edge.destination.other_end
    leg = head_edge.destination.leg
    inside_path = []
    e = head_edge.next_edge
    while e != leg:
        inside_path += [e.twin]
        e = e.next_edge
    e = e.twin.next_edge
    while e != tail_edge:
        inside_path += [e.twin]
        e = e.next_edge
    return inside_path

def prepareInsideFacesOrder(G, corners, inside_path):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e inside_path sendo uma lista de semi-arestas formando um
    caminho
    
    Inicializa as arestas do contorno de cada face anexada a inside_path para
    que sejam numeradas de acordo com a ordem que aparecem em inside_path
    """
    ip = inside_path
    for k in range(len(ip)): ip[k].order = -1
    k = 0
    cleanFaceOrder(G, corners, ip[k])
    k += 1
    while k < len(ip):
        if ip[k].order == -1: cleanFaceOrder(G, corners, ip[k])
        k += 1

def findMaximumLeg(G, corners, first, order):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, uma semi-aresta first e um número order representando 
    
    Sendo f a face incidente a first, existem order - 1 faces anexadas a um
    caminho e que foram processadas anteriormene
    
    Devolve leg_value e leg, sendo que leg é a perna de um ciclo crítico maximal
    contido no contorno da f se leg_value > 1
    """
    leg_value, leg = 0, None
    first.order = order
    first.twin.order = order
    e = first.next_edge
    while e != first:
        if e.order is None: e.twin.order = order
        e.order = order
        value = e.order - e.twin.order
        if value > leg_value:
            leg = e.twin
            leg_value = value
        e = e.next_edge
    return leg_value, leg

def findThreeLegs(G, corners, inner_leg, outer_leg):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H e duas pernas inner_leg e outer_leg de um ciclo C'
    
    C' deve ser um ciclo crítico maximal anexado a Q_H(C), com n_{AH}(C') = 1,
    onde C é um ciclo crítico maximal anexado ao NS-caminho mais à esquerda
    P satisfazendo n_H(C) = 1
    
    outer_leg deve ser a perna horária de C e inner_leg a perna anti-horária de C'
    
    Devolve success, certificate, True e cycle_legs
    
    Se outer_leg é perna de C', devolve success = True, certificate = None e
    cycle_legs sendo uma tupla de pernas (head_edge, inner_leg, tail_edge) onde:
    - head_edge é a perna incidente ao vértice anterior de C'
    - tail_edge é a perna incidente ao vértice posterior de C'
    Caso contrário, devolve success = False, certificate = None e cycle_legs = None
    """
    e = inner_leg
    while e.order != e.twin.order: e = e.next_edge
    head_edge = e
    e = inner_leg.twin
    while e.order != e.twin.order: e = e.prev_edge
    tail_edge = e
    cycle_legs = (head_edge, inner_leg, tail_edge)
    e = head_edge.twin
    while e.twin != tail_edge:
        if e == outer_leg: return True, None, True, cycle_legs
        e = e.next_edge
    return False, BadCycleInside(G, inner_leg, outer_leg, tail_edge), True, None

def partitionSmallCycle(G, corners, memo, head_edge, cycle_legs):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    de um subgrafo H, a tupla (a_list, b_list, p) e as pernas head_edge e 
    cycle_legs
    
    head_edge deve ser a perna de um ciclo crítico maximal C anexado ao NS-caminho
    mais à esquerda P incidente ao seu vértice anterior
    
    cycle_legs deve ser uma tupla de pernas (head_edge, inner_leg, tail_edge) onde:
    - head_edge é a perna incidente ao vértice anterior de C'
    - inner_leg é a perna anti-horária de C'
    - tail_edge é a perna incidente ao vértice posterior de C'
    sendo C' um ciclo crítico maximal anexado a Q_H(C)
    
    Particiona C de acordo com o caso em que n_H(C) = 1 e n_{AH}(C) > 1 e C tem
    um ciclo crítico anti-horário C' anexado a Q_H(C)
    """
    
    leg = head_edge.destination.leg
    tail_edge = head_edge.destination.other_end
    e = head_edge.next_edge
    while e.twin != cycle_legs[0]:
        setDirection(e, "vertical", "W", "E")
        e = e.next_edge
    setDirection(e, "vertical", "W", "E")
    new_corners = [e.destination]
    # lado horário
    e = e.next_edge
    while e != leg:
        setDirection(e, "horizontal", "W", "N")
        e = e.next_edge
    new_corners += [e.origin]
    e = e.twin.next_edge
    while e.twin != cycle_legs[2]:
        setDirection(e, "vertical", "W", "E")
        e = e.next_edge
    new_corners += [e.origin]
    # lado anti-horário
    e = e.twin.next_edge
    while e.twin != cycle_legs[1]:
        setDirection(e, "horizontal", "E", "S")
        e = e.next_edge
    new_corners += [e.origin]
    e = e.twin.next_edge
    while e != cycle_legs[0]:
        setDirection(e, "vertical", "E", "W")
        e = e.next_edge
    e = cycle_legs[2].twin
    # lado horário
    while e != tail_edge:
        setDirection(e, "vertical", "W", "E")
        e = e.next_edge
    return tuple(new_corners)

def _createMidFrame(G, corners, p, separators):
    global FRAME_COUNTER
    edge_texts = G.e*[None]
    edge_colors = G.e*[0]
    edge_styles = G.e*['-']
    if p:
        for e in p:
            edge_colors[e.index[0]] = 1
            edge_styles[e.index[0]] = '--'
    for e in G.edges:
        if e[0].direction:
            edge_texts[e[0].index[0]] = e[0].direction[0].upper()
            edge_colors[e[0].index[0]] = 2
            edge_styles[e[0].index[0]] = '-'
    first = StartingHalfEdge(G, corners, "N")
    edge_colors[first.index[0]] = 3
    edge_texts[first.index[0]] = first.label[0].upper()
    edge_styles[first.index[0]] = '='
    e = first.outer_next
    while e != first:
        edge_colors[e.index[0]] = 3
        edge_texts[e.index[0]] = e.label[0].upper()
        edge_styles[first.index[0]] = '='
        e = e.outer_next
    faces = [f.index for f in separators]
    _customizedPlot(G, G.v*[None], edge_texts, edge_colors, edge_styles, faces)
    _saveFrame(PATH, FRAME_COUNTER)
    FRAME_COUNTER += 1

def _createLastFrame(G, success, certificate):
    global FRAME_COUNTER
    edge_texts, edge_colors = [], []
    if success:
        edge_texts = [e[0].direction[0].upper() for e in G.edges]
        edge_colors = G.e*[2]
    else:
        edge_texts = G.e*[None]
        edge_colors = G.e*[0]
        for e in certificate: edge_colors[e.index[0]] = 2
    _customizedPlot(G, G.v*[None], edge_texts, edge_colors, G.e*['-'], [])
    _saveFrame(PATH, FRAME_COUNTER)
    FRAME_COUNTER += 1
