#!/usr/bin/python3

from Plotters import _prepareForFrames, _customizedPlot, _saveFrame, _createAnimation

# VARIÁVEIS GLOBAIS
ANIMATED = False
FRAME_COUNTER = 0
PATH = 'grade'

def AnimatedHorizontalSpanningTree(G, corners, direction, path='grade', fps=8, figsize=(4,4)):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO, uma lista indicando as direções de cada
    aresta de G, uma string path incidando um caminho para um diretório existente
    e um número indicando o FPS (frames por segundo)
    
    Devolve o resultado de HorizontalSpanningTree(G, corners, direction)
    
    Salva um .gif no caminho e nome indicado por path e as figuras dos frames
    gerados pelo desenho T_y no diretório path.
    
    Exemplo:
        path = '/home/user/output'
    Resultado:
        '/home/user/output.gif' e
        '/home/user/output/0.png', .., '/home/user/output/20.png'
    """
    global ANIMATED, PATH, FRAME_COUNTER
    hasCoords = True
    for v in G.vertices:
        if type(v.x) != int and type(v.x) != float: hasCoords = False
        if type(v.y) != int and type(v.y) != float: hasCoords = False
    if not hasCoords:
        raise Exception("G has vertices without coordinates!")
    ANIMATED = True
    PATH = path
    FRAME_COUNTER = 0
    _prepareForFrames(path, figsize)
    result = HorizontalSpanningTree(G, corners, direction)
    _createLastFrame(G, corners, direction)
    _createAnimation(path, fps)
    ANIMATED = False
    return result

def AnimatedVerticalSpanningTree(G, corners, direction, path='grade', fps=8, figsize=(4,4)):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO, uma lista indicando as direções de cada
    aresta de G, uma string path incidando um caminho para um diretório existente
    e um número indicando o FPS (frames por segundo)
    
    Devolve o resultado de VerticalSpanningTree(G, corners, direction)
    
    Salva um .gif no caminho e nome indicado por path e as figuras dos frames
    gerados pelo desenho de T_x no diretório path.
    
    Exemplo:
        path = '/home/user/output'
    Resultado:
        '/home/user/output.gif' e
        '/home/user/output/0.png', .., '/home/user/output/20.png'
    """
    global ANIMATED, PATH, FRAME_COUNTER
    hasCoords = True
    for v in G.vertices:
        if type(v.x) != int and type(v.x) != float: hasCoords = False
        if type(v.y) != int and type(v.y) != float: hasCoords = False
    if not hasCoords:
        raise Exception("G has vertices without coordinates!")
    ANIMATED = True
    PATH = path
    FRAME_COUNTER = 0
    _prepareForFrames(path, figsize)
    result = VerticalSpanningTree(G, corners, direction)
    _createLastFrame(G, corners, direction)
    _createAnimation(path, fps)
    ANIMATED = False
    return result

def HorizontalSpanningTree(G, corners, direction):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO e uma lista indicando as direções de cada
    aresta de G
    
    Devolve 2 listas:
    - uma lista indicando para cada aresta e de G, se e está em T_y
    (se e está em T_y, thickness associado a e é "thick", caso contrário, é "dotted")
    - uma lista com a y-coordenada de cada vértice de G
    
    A construção de tais listas é feita por uma busca em profundidade anti-horária
    a partir da aresta saindo de NO que aponta para baixo
    """
    for e in G.edges: e[0].thickness = None
    for v in G.vertices:
        v.visited = False
        v.temp = 0
    e = corners[0].incident_edge
    while e.incident_face == G.outer_face: e = e.prev_edge.twin
    HorizontalSpanningTreeRecur(G, corners, e, direction, "thick")
    HorizontalSpanningTreeRecur(G, corners, e.prev_edge.twin, direction, "thick")
    return [e[0].thickness for e in G.edges], [v.temp for v in G.vertices]

def HorizontalSpanningTreeRecur(G, corners, first, direction, thickness):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO, uma lista indicando as direções de cada
    aresta de G, uma semi-aresta first ainda não percorrida na busca em profundidade,
    uma lista indicando as direções de cada aresta de G e thickness indicando se
    first faz parte de T_y
    
    Continua a busca em profundidade a partir de first (considere v o vértice da
    origem de first):
    Se first é horizontal, passa o y temporário para os próximos vértices e, ao
    fim da execução, atribui a y-coordenada final de v
    Se first é vertical, caso first não seja uma semi-aresta de T_y, atribui
    o y temporário para o vértice u acima de v e não continua a busca a partir de u
    Caso first seja uma semi-aresta de T_y, atribui a y-coordenada final de v ao fim
    da execução
    """
    path = PassPathBetweenVertices(G, corners, first, thickness)
    first, last = path[0], path[-1]
        
    if thickness == "dotted":
        PassDottedPath(G, path)
        if ANIMATED: _createMidFrame(G, corners, direction, path)
        return

    if direction[first.index[0]] == "horizontal":
        last.destination.temp = max(first.origin.temp, last.destination.temp)
    
    if ANIMATED: _createMidFrame(G, corners, direction, path)
    
    edges = EdgesToVisit(G, last)    
    for k in range(len(edges)):
        e = edges[k]
        if direction[e.index[0]] == "horizontal":
            HorizontalSpanningTreeRecur(G, corners, e, direction, "thick")
        elif k == 1 or e.origin == corners[2]:
            HorizontalSpanningTreeRecur(G, corners, e, direction, "dotted")
        elif not e.origin.visited:
            HorizontalSpanningTreeRecur(G, corners, e, direction, "thick")
        
    if direction[first.index[0]] == "horizontal":
        y = last.destination.temp
        for e in path: e.origin.temp = y
    else:
        PassReversedPath(G, path)

    if ANIMATED: _createMidFrame(G, corners, direction, path)

    return

def VerticalSpanningTree(G, corners, direction):
    """
    Recebe um PlaneGraph G e uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO e uma lista indicando as direções de cada
    aresta de G
    
    Devolve 2 listas:
    - uma lista indicando para cada aresta e de G, se e está em T_x
    (se e está em T_x, thickness associado a e é "thick", caso contrário, é "dotted")
    - uma lista com a x-coordenada de cada vértice de G
    
    A construção de tais listas é feita por uma busca em profundidade anti-horária
    a partir da aresta saindo de SO que aponta para a direita
    """
    for e in G.edges: e[0].thickness = None
    for v in G.vertices:
        v.visited = False
        v.temp = 0
    e = corners[1].incident_edge
    while e.incident_face == G.outer_face: e = e.prev_edge.twin
    VerticalSpanningTreeRecur(G, corners, e, direction, "thick")
    VerticalSpanningTreeRecur(G, corners, e.prev_edge.twin, direction, "thick")
    return [e[0].thickness for e in G.edges], [v.temp for v in G.vertices]

def VerticalSpanningTreeRecur(G, corners, first, direction, thickness):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO, uma lista indicando as direções de cada
    aresta de G, uma semi-aresta first ainda não percorrida na busca em profundidade,
    uma lista indicando as direções de cada aresta de G e thickness indicando se
    first faz parte de T_x
    
    Continua a busca em profundidade a partir de first (considere v o vértice da
    origem de first):
    Se first é vertical, passa o x temporário para os próximos vértices e, ao
    fim da execução, atribui a x-coordenada final de v
    Se first é horizontal, caso first não seja uma semi-aresta de T_x, atribui
    o x temporário para o vértice u à direita de v e não continua a busca a partir de u
    Caso first seja uma semi-aresta de T_x, atribui a x-coordenada final de v ao fim
    da execução
    """
    path = PassPathBetweenVertices(G, corners, first, thickness)
    first, last = path[0], path[-1]

    if thickness == "dotted":
        PassDottedPath(G, path)
        if ANIMATED: _createMidFrame(G, corners, direction, path)
        return

    if direction[first.index[0]] == "vertical":
        last.destination.temp = max(first.origin.temp, last.destination.temp)
    
    if ANIMATED: _createMidFrame(G, corners, direction, path)

    edges = EdgesToVisit(G, last)
    for k in range(len(edges)):
        e = edges[k]
        if direction[e.index[0]] == "vertical":
            VerticalSpanningTreeRecur(G, corners, e, direction, "thick")
        elif k == 1 or e.origin == corners[3]:
            VerticalSpanningTreeRecur(G, corners, e, direction, "dotted")
        elif not e.origin.visited:
            VerticalSpanningTreeRecur(G, corners, e, direction, "thick")

    if direction[first.index[0]] == "vertical":
        y = last.destination.temp
        for e in path: e.origin.temp = y
    else:
        PassReversedPath(G, path)
    
    if ANIMATED: _createMidFrame(G, corners, direction, path)
    
    return

def PassPathBetweenVertices(G, corners, first, thickness):
    """
    Recebe um PlaneGraph G, uma tupla corners contendo os 4 vértices de cantos
    em ordem horária a partir de NO, uma lista indicando as direções de cada aresta de G
    e thickness indicando se first faz parte de T_x
    
    Devolve uma lista de semi-arestas path formando o menor caminho contendo
    first tal que as pontas sejam vértices de grau 3 ou cantos
    """
    path = [first]
    e = first
    while e.destination.degree != 3 and e.destination not in corners:
        e.thickness = thickness
        e = e.next_edge
        path += [e]
    e.thickness = thickness
    return path

def PassDottedPath(G, path):
    """
    Recebe um PlaneGraph G e uma lista de semi-arestas formando um caminho path

    Atualiza as coordenadas temporárias de path considerando que as arestas de
    path não estão em T_t, com t sendo x ou y
    """
    t = path[0].origin.temp
    k = 0
    for e in path:
        e.origin.temp = t + k
        k += 1
    path[-1].destination.temp = t + k
    path[-1].destination.visited = True

def EdgesToVisit(G, visited_e):
    """
    Recebe um PlaneGraph G e visited_e como sendo a última aresta percorrida na
    busca em profundidade

    Devolve uma lista com as próximas arestas saindo de visited_e.destination
    em ordem anti-horária
    """
    e = visited_e.twin.prev_edge
    edges = []
    while e != visited_e:
        edges += [e.twin]
        e = e.twin.prev_edge
    return edges

def PassReversedPath(G, path):
    """
    Recebe um PlaneGraph G e uma lista de semi-arestas formando um caminho path

    Atribui as coordenadas finais para os vértices de path considerando que as
    arestas de path estão no sentido em que t decresce, com t sendo x ou y
    """
    t = path[-1].destination.temp
    k = 1
    for e in reversed(path[1:]):
        e.origin.temp = t + k
        k += 1
    path[0].origin.temp = max(path[0].origin.temp, t + k)

def _createMidFrame(G, corners, direction, path):
    global FRAME_COUNTER
    vertex_texts = [v.temp for v in G.vertices]
    edge_texts = [d[0].upper() for d in direction]
    edge_colors = G.e*[0]
    edge_styles = G.e*['-']
    for e in G.edges:
        if e[0].thickness:
            if e[0].thickness == "thick": color, style = 2, '-'
            else: color, style = 1, '--'
            edge_colors[e[0].index[0]] = color
            edge_styles[e[0].index[0]] = style
    if path[0].thickness == "thick":
        for e in path:
            edge_colors[e.index[0]] = 3 
            edge_styles[e.index[0]] = '='
    _customizedPlot(G, vertex_texts, edge_texts, edge_colors, edge_styles, [], True)
    _saveFrame(PATH, FRAME_COUNTER)
    FRAME_COUNTER += 1

def _createLastFrame(G, corners, direction):
    global FRAME_COUNTER
    vertex_texts = [v.temp for v in G.vertices]
    edge_texts = [d[0].upper() for d in direction]
    edge_colors, edge_styles = [], []
    for e in G.edges:
        if e[0].thickness == "thick": color, style = 2, '-'
        else: color, style = 0, ''
        edge_colors += [color]
        edge_styles += [style]
    _customizedPlot(G, vertex_texts, edge_texts, edge_colors, edge_styles, [], True)
    _saveFrame(PATH, FRAME_COUNTER)
    FRAME_COUNTER += 1
