#!/usr/bin/python3

from math import floor, ceil, cos, sin, pi, sqrt

def intoargs(infile):
    """
    Recebe um arquivo infile no padrão .in e devolve vertices, edges e faces 
    para servirem de argumento do construtor de PlaneGraph
    
    O arquivo .in segue o seguinte padrão:
    V
    [nomev_0] [x_0 y_0]
    ...
    [nomev_{V-1}] [x_{V-1} y_{V-1}]
    A
    [nomea_0] u_0 v_0
    ...
    [nomea_{A-1}] u_{A-1} v_{A-1}
    F
    [nomef_0] o_0 a_{0,0} a_{0,1} ... a_{0,n_0}
    ...
    [nomef_{F-1}] o_{F-1} a_{F-1,0} a_{F-1,1} ... a_{F-1,n_{F-1}}
    
    onde V = número de vértices, A = número de arestas, F = número de faces,
    sendo que A está V+1 linhas após V e F está A+1 linhas após A.
    Para cada vértice, aresta ou face, podemos opcionalmente atribuir um nome.
    Para cada vértice, também podemos opcionalmente atribuir coordenadas.
    Para cada aresta de índice k, k = 0..A-1, u_k e v_k são os índices dos
    vértices de suas pontas. (u_k, v_k) será considerado o sentido padrão da
    aresta.
    Para cada face de índice k, k = 0..F-1, o_k pode ser + ou - e representa se
    a orientação da aresta a_{k,0} deve ser tomado em sentido padrão ou contrário,
    respectivamente, ao percorrermos a face em sentido anti-horário. As arestas
    a_{k,0}, a_{k,1}, ..., a_{k,n_k} formam o ciclo correspondente ao contorno
    da face de índice k.
    A primeira face, de índice 0, deve necessariamente ser a face exterior.
    
    Exemplo (triângulo):
    
    3
    
    X -0.5000 -0.8660
    -0.5000 0.8660
    3
    0 1
    1 2
    2 0
    2
    C_o + 0 1 2
    - 2 1 0
    
    """
    v = int(infile.readline())
    vertices = []
    for k in range(v):
        param = infile.readline().strip().split()
        l = len(param)
        name, coords = None, None
        if l % 2 == 1: name = param[0]
        if l >= 2: coords = (float(param[l//3]), float(param[l//3 + 1]))
        vertices += [(name, coords)]

    e = int(infile.readline())
    edges = []
    for k in range(e):
        param = infile.readline().strip().split()
        l = len(param)
        name = param[0] if l % 2 == 1 else None
        vindices = (int(param[l//3]), int(param[l//3 + 1]))
        edges += [(name, vindices)]

    f = int(infile.readline())
    faces = []
    for k in range(f):
        param = infile.readline().strip().split()
        l = len(param)
        if param[1] in ["+", "-"]:
            name = param[0]
            orientation = param[1]
            eindices = [int(p) for p in param[2:]]
        else:
            name = None
            orientation = param[0]
            eindices = [int(p) for p in param[1:]]
        faces += [(name, orientation, eindices)]

    return vertices, edges, faces

def _matrixtograph(matrix, m, n):
    """
    Cria listas vertices, edges para servirem de argumento ao construtor de PlaneGraph
    Também cria lista de dicionários adjacency: dado um par (i, j) se o vértice
    de índice i e o vértice de índice j são adjacentes, adjacency[i][j] é (k, o)
    onde k é o índice da aresta entre i e j e o é a orientação
    
    Ao final da execução, todos os '+' de matrix estarão sobrescritos com índices
    dos vértices correspondentes.
    """
    matrix[0][0] = 0
    vertices = [(None, (0, 0))]
    edges = []
    adjacency = [{}]
    stack = [(0,0)]
    while stack:
        i, j = stack.pop()
        k = j+1
        while k < n and matrix[i][k] == '-': k += 1
        if k < n and matrix[i][k] != ' ':
            if matrix[i][k] == '+':
                matrix[i][k] = len(vertices)
                vertices += [(None, (k, -i))]
                adjacency += [{}]
                stack.append((i, k))
            adjacency[matrix[i][j]][matrix[i][k]] = (len(edges), '+')
            adjacency[matrix[i][k]][matrix[i][j]] = (len(edges), '-')
            edges += [(None, (matrix[i][j], matrix[i][k]))]
        k = i+1
        while k < m and matrix[k][j] == '|': k += 1
        if k < m and matrix[k][j] != ' ':
            if matrix[k][j] == '+':
                matrix[k][j] = len(vertices)
                vertices += [(None, (j, -k))]
                adjacency += [{}]
                stack.append((k, j))
            adjacency[matrix[i][j]][matrix[k][j]] = (len(edges), '+')
            adjacency[matrix[k][j]][matrix[i][j]] = (len(edges), '-')
            edges += [(None, (matrix[i][j], matrix[k][j]))]
    return vertices, edges, adjacency

def _matrixtoouterface(matrix, m, n, adjacency):
    """
    Devolve os parâmetros da face exterior: uma lista boundary contendo as arestas de
    seu contorno e orientation indicando se a primeira aresta deve ser percorrida
    no sentido padrão ('+') ou não ('-').
    """
    orientation = None
    boundary = []
    vindex = matrix[0][0]
    i, j = 0, 1
    while j < n:
        while matrix[i][j] == '-': j += 1
        he = adjacency[vindex][matrix[i][j]]
        if orientation is None: orientation = he[1]
        boundary += [he[0]]
        vindex = matrix[i][j]
        j += 1
    i, j = 1, n-1
    while i < m:
        while matrix[i][j] == '|': i += 1
        he = adjacency[vindex][matrix[i][j]]
        boundary += [he[0]]
        vindex = matrix[i][j]
        i += 1
    i, j = m-1, n-2
    while j >= 0:
        while matrix[i][j] == '-': j -= 1
        he = adjacency[vindex][matrix[i][j]]
        boundary += [he[0]]
        vindex = matrix[i][j]
        j -= 1
    i, j = m-2, 0
    while i >= 0:
        while matrix[i][j] == '|': i -= 1
        he = adjacency[vindex][matrix[i][j]]
        boundary += [he[0]]
        vindex = matrix[i][j]
        i -= 1
    return (None, orientation, boundary)

def _matrixtoinnerface(matrix, m, n, adjacency, i, j):
    """
    Devolve os parâmetros da face que tem o vértice em matrix[i][j] como canto
    superior esquerdo: uma lista boundary contendo as arestas de seu contorno e
    orientation indicando se a primeira aresta deve ser percorrida no sentido
    padrão ('+') ou não ('-').
    """
    orientation = None
    boundary = []
    vindex = matrix[i][j]
    i += 1
    while i < m:
        while matrix[i][j] == '|': i += 1
        if matrix[i][j] == ' ': break
        he = adjacency[vindex][matrix[i][j]]
        if orientation is None: orientation = he[1]
        boundary += [he[0]]
        vindex = matrix[i][j]
        if matrix[i][j+1] == '-': break
        i += 1
    if i >= m: i -= 1
    j += 1
    while j < n:
        while matrix[i][j] == '-': j += 1
        if matrix[i][j] == ' ': break
        he = adjacency[vindex][matrix[i][j]]
        boundary += [he[0]]
        vindex = matrix[i][j]
        if matrix[i-1][j] == '|': break
        j += 1
    if j >= n: j -= 1
    i -= 1
    while i >= 0:
        while matrix[i][j] == '|': i -= 1
        if matrix[i][j] == ' ': break
        he = adjacency[vindex][matrix[i][j]]
        boundary += [he[0]]
        vindex = matrix[i][j]
        if matrix[i][j-1] == '-': break
        i -= 1
    if i < 0: i += 1
    j -= 1
    while j >= 0:
        while matrix[i][j] == '-': j -= 1
        if matrix[i][j] == ' ': break
        he = adjacency[vindex][matrix[i][j]]
        boundary += [he[0]]
        vindex = matrix[i][j]
        if matrix[i+1][j] == '|': break
        j -= 1
    return (None, orientation, boundary)

def matrixtoargs(mtxfile):
    """
    Recebe um arquivo mtxfile no padrão .mtx e devolve vertices, edges e faces 
    para servirem de argumento do construtor de PlaneGraph
    
    Complexidade: Theta(HW)
    
    O arquivo .mtx segue o seguinte padrão:
    H W
    (desenho retangular em ASCII com altura H e largura W)
    
    Exemplo:
    
    21 19
    +-+-------+-+-+---+
    | |       | | |   |
    | +-+-+-+-+ | |   |
    | | | | | | | |   |
    | | +-+ | | | +-+-+
    | | | | | | | | | |
    +-+-+-+-+ | | | | |
    |       | | | | | |
    +-------+ +-+-+ | |
    |       | |   | | |
    |       +-+-+-+-+-+
    |       |   |     |
    +-+-----+   |     |
    | |     |   |     |
    | +-+-+-+   +---+-+
    | | | | |   |   | |
    +-+ +-+-+-+-+---+ |
    | | |     |     | |
    +-+-+     |     | |
    |   |     |     | |
    +-+-+-----+-----+-+
    
    No desenho, as arestas verticais devem ser representadas pelo caractere |,
    as horizontais por - e cada vértice por +.
    
    """
    param = mtxfile.readline().strip().split()
    m, n = int(param[0]), int(param[1])
    matrix = [list(mtxfile.readline().strip()) for k in range(m)]

    vertices, edges, adjacency = _matrixtograph(matrix, m, n)
    faces = [_matrixtoouterface(matrix, m, n, adjacency)]
    for v in vertices:
        i, j = -v[1][1], v[1][0]
        if i+1 < m and matrix[i+1][j] == '|':
            if j+1 < n and matrix [i][j+1] == '-':
                faces += [_matrixtoinnerface(matrix, m, n, adjacency, i, j)]

    return vertices, edges, faces

def unityRoot(n):
    """
    Devolve uma função f tal que f(k) é a k-ésima potência da n-ésima raiz da unidade
    """
    def rootPower(k):
        k = k % n
        return (cos(2*k*pi/n), sin(2*k*pi/n))
    return rootPower

def convexCombination(p1, p2, r):
    """
    Devolve a combinação convexa (1-r)*p1 + r*p2
    """
    return (p1[0]*(1-r) + p2[0]*r, p1[1]*(1-r) + p2[1]*r)

def segment(p1, p2):
    """
    Devolve uma função f tal que f(n)(k) é o k-ésimo ponto do segmento entre
    p1 e p2 subdividido em n+1 partes, i.e., subdividido de modo que existam
    n pontos entre p1 e p2
    """
    def subdividedSegment(n):
        def segmentPoint(k):
            r = k / (n+1)
            return convexCombination(p1, p2, r)
        return segmentPoint
    return subdividedSegment

def coordsBetween(l1, l2, r):
    """
    Devolve uma função f tal que f(n)(k) é o k-ésimo ponto do segmento entre
    p1 e p2 subdividido em n+1 partes, i.e., subdividido de modo que existam
    n pontos entre p1 e p2
    """
    def subdividedSegment(n):
        def segmentPoint(k):
            l, m = len(l1), len(l2)
            r1 = k * (l-1) / (n+1)
            r2 = k * (m-1) / (n+1)
            p1 = convexCombination(l1[floor(r1)], l1[ceil(r1)], r1 - floor(r1))
            p2 = convexCombination(l2[floor(r2)], l2[ceil(r2)], r2 - floor(r2))
            return convexCombination(p1, p2, r)
        return segmentPoint
    return subdividedSegment

def distance2(p1, p2):
    """
    Devolve o quadrado da distância euclidiana entre p1 e p2
    """
    return (p1[0] - p2[0])**2 + (p1[1] - p2[1])**2

def distance(p1, p2):
    """
    Devolve a distância euclidiana entre p1 e p2
    """
    return sqrt(distance2(p1, p2))

def normalizedVector(p1, p2):
    """
    Devolve o vetor p2 - p1 normalizado
    """
    norm = distance(p1, p2)
    return [(p2[0] - p1[0])/norm, (p2[1] - p1[1])/norm]

# convexity of p2 - p1 and p3 - p1 (counterclockwise angle)
def convex(p1, p2, p3):
    """
    Devolve se o ângulo formado por p2 p1 p3 é menor ou igual a pi rad
    
    Obs.: Um polígono convexo não tem um ângulo interno maior que pi rad
    """
    return (p2[0] - p1[0]) * (p3[1] - p1[1]) - (p3[0] - p1[0]) * (p2[1] - p1[1])  >= 0
